"""Managers for reviewboard.reviews.models."""

from __future__ import annotations

import logging
from typing import Dict, Optional, TYPE_CHECKING, Union

from django.contrib.auth.models import AnonymousUser, User
from django.core.exceptions import ObjectDoesNotExist
from django.db import connections, router, transaction, IntegrityError
from django.db.models import Exists, Manager, OuterRef, Q
from django.db.models.query import QuerySet
from django.utils.text import slugify
from djblets.db.managers import ConcurrencyManager
from housekeeping.functions import deprecate_non_keyword_only_args

from reviewboard.deprecation import RemovedInReviewBoard90Warning
from reviewboard.diffviewer.models import DiffSetHistory
from reviewboard.reviews.signals import review_request_diffset_uploaded
from reviewboard.scmtools.errors import ChangeNumberInUseError
from reviewboard.scmtools.models import Repository
from reviewboard.site.models import LocalSite

if TYPE_CHECKING:
    from reviewboard.attachments.models import FileAttachment
    from reviewboard.changedescs.models import ChangeDescription
    from reviewboard.integrations.base import Integration
    from reviewboard.integrations.models import IntegrationConfig
    from reviewboard.reviews.models import (
        FileAttachmentComment,
        Review,
        ReviewRequest,
        StatusUpdate,
    )
    from reviewboard.site.models import AnyOrAllLocalSites


logger = logging.getLogger(__name__)


class CommentManager(ConcurrencyManager):
    """A manager for Comment models.

    This handles concurrency issues with Comment models.

    Version Added:
        5.0
    """

    def accessible(self, user, extra_query=None, local_site=None,
                   distinct=False):
        """Return a queryset for comments accessible by the given user.

        For superusers, all comments in all reviews will be returned.

        For regular users, only comments in reviews that they own or that
        are in the repositories, local sites, and review groups which the
        user has access to will be returned.

        For anonymous users, only comments that are in public repositories
        and whose review requests are not targeted by invite-only
        review groups will be returned.

        Args:
            user (django.contrib.auth.models.User):
                The User object that must have access to any
                returned comments.

            extra_query (django.db.models.Q, optional):
                Additional query parameters to add for filtering
                down the resulting queryset.

            local_site (reviewboard.site.models.LocalSite or
                        reviewboard.site.models.LocalSite.ALL, optional):
                A specific :term:`Local Site` that the comments must be
                associated with. It is assumed that the given user has access
                to the :term:`Local Site`. By default, this will only return
                comments not part of a site.

                This may be :py:attr:`LocalSite.ALL
                <reviewboard.site.models.LocalSite.ALL>`.

            distinct (bool, optional):
                Whether to return distinct results.

                Turning this on can decrease performance. It's off by default.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.
        """
        assert isinstance(user, (User, AnonymousUser))

        return self._query(user=user,
                           public=None,
                           extra_query=extra_query,
                           local_site=local_site,
                           status=None,
                           filter_private=True,
                           distinct=distinct)

    def from_user(self, user, *args, **kwargs):
        """Return the query for comments created by a user.

        Args:
            user (django.contrib.auth.models.User):
                The User object to query for.

            *args (tuple):
                Additional positional arguments to pass to the common
                :py:meth:`_query` function.

            **kwargs (dict):
                Additional keyword arguments to pass to the common
                :py:meth:`_query` function.

        Returns:
            django.db.models.query.QuerySet:
            A queryset for all the comments created by the given user.
        """
        assert isinstance(user, User)

        return self._query(extra_query=Q(review__user=user), *args, **kwargs)

    def _query(self, user=None, public=None, status='P', extra_query=None,
               local_site=None, filter_private=False, distinct=False):
        """Do a query for comments.

        Args:
            user (django.contrib.auth.models.User, optional):
                A user to query for.

            public (bool or None, optional):
                Whether to filter for comments from public (published) reviews.
                If set to `None`, comments from both published and unpublished
                reviews will be included.

            status (unicode, optional):
                The status of the review request that comments are associated
                with.

            extra_query (django.db.models.Q, optional):
                Additional query parameters to add.

            local_site (reviewboard.site.models.LocalSite or
                        reviewboard.site.models.LocalSite.ALL, optional):
                A local site to limit to, if appropriate. If a user is given,
                it is assumed that they have access to the :term:`Local Site`.
                By default, this will only return comments not part of a site.

                This may be :py:attr:`LocalSite.ALL
                <reviewboard.site.models.LocalSite.ALL>`.

            filter_private (bool, optional):
                Whether to filter out comments from review requests on private
                repositories or invite-only review groups that the user
                does not have access to. This will also filter out comments
                from unpublished reviews that are not owned by the user.

                This requires ``user`` to be provided.

            distinct (bool, optional):
                Whether to return distinct results.

                Turning this off can increase performance. It's on by default
                for backwards-compatibility.

        Returns:
            django.db.models.query.QuerySet:
            A queryset for the given conditions.
        """
        from reviewboard.reviews.models import Group

        q = Q()

        if local_site is not LocalSite.ALL:
            q &= Q(review__review_request__local_site=local_site)

        if status:
            q &= Q(review__review_request__status=status)

        if extra_query:
            q &= extra_query

        if filter_private and (not user or not user.is_superuser):
            repo_q = Q(review__review_request__repository=None)
            group_q = Q(review__review_request__target_groups=None)

            # TODO: should be consolidated with queries in ReviewRequestManager
            if user and user.is_authenticated:
                accessible_repo_ids = Repository.objects.accessible_ids(
                    user=user,
                    visible_only=False,
                    local_site=local_site)
                accessible_group_ids = Group.objects.accessible_ids(
                    user=user,
                    visible_only=False,
                    local_site=local_site)

                repo_q |= Q(('review__review_request__repository__in',
                             accessible_repo_ids))
                group_q |= Q(('review__review_request__target_groups__in',
                              accessible_group_ids))

                acl_check_q = (
                    repo_q &
                    (Q(review__review_request__target_people=user) |
                     group_q)
                )

                if public is None:
                    q &= (
                        Q(review__user=user) |
                        (Q(review__public=True) &
                         acl_check_q)
                    )
                elif public:
                    q &= Q(review__public=True) & acl_check_q
                else:
                    q &= Q(review__public=False) & Q(review__user=user)
            else:
                # Return an empty result when an unauthenticated user queries
                # for comments from unpublished reviews.
                if public is False:
                    return self.none()

                repo_q |= Q(review__review_request__repository__public=True)
                group_q |= \
                    Q(review__review_request__target_groups__invite_only=False)

                q &= repo_q & group_q & Q(review__public=True)
        else:
            if public is not None:
                q &= Q(review__public=public)

        queryset = self.filter(q)

        if distinct:
            queryset = queryset.distinct()

        return queryset


class FileAttachmentCommentManager(CommentManager):
    """A manager for FileAttachmentComment models.

    Version Added:
        8.0
    """

    def for_file_attachment(
        self,
        *,
        attachment: FileAttachment | None,
        diff_against_file_attachment: (FileAttachment | None) = None,
        user: (User | AnonymousUser | None) = None,
    ) -> QuerySet[FileAttachmentComment]:
        """Return a queryset for accessible comments on a given attachment.

        Args:
            attachment (reviewboard.attachments.models.FileAttachment):
                The file attachment.

            diff_against_file_attachment (reviewboard.attachments.models.
                                          FileAttachment, optional):
                The file attachment being diffed against.

            user (django.contrib.auth.models.User, optional):
                The user making the request.

        Returns:
            django.db.models.QuerySet:
            A queryset for the visible comments.

        Raises:
            ValueError:
                The method was called without a valid file attachment.
        """
        if attachment is None and diff_against_file_attachment is None:
            raise ValueError(
                'attachment and diff_against_file_attachment cannot both be '
                'None'
            )

        q = Q(file_attachment=attachment)

        if diff_against_file_attachment is not None:
            q &= Q(diff_against_file_attachment=diff_against_file_attachment)

        if user is not None and user.is_authenticated:
            q &= Q(review__public=True) | Q(review__user=user)
        else:
            q &= Q(review__public=True)

        return self.filter(q)


class DefaultReviewerManager(Manager):
    """A manager for DefaultReviewer models."""

    def for_repository(self, repository, local_site):
        """Returns all DefaultReviewers that represent a repository.

        These include both DefaultReviewers that have no repositories
        (for backwards-compatibility) and DefaultReviewers that are
        associated with the given repository.
        """
        return self.filter(local_site=local_site).filter(
            Q(repository__isnull=True) | Q(repository=repository))

    def can_create(self, user, local_site=None):
        """Returns whether the user can create default reviewers."""
        return (user.is_superuser or
                (local_site and local_site.is_mutable_by(user)))


class ReviewGroupManager(Manager):
    """A manager for Group models."""

    @deprecate_non_keyword_only_args(RemovedInReviewBoard90Warning)
    def accessible(self,
                   user,
                   *,
                   visible_only=True,
                   local_site=None,
                   distinct=True):
        """Return a queryset for review groups accessible by the given user.

        For superusers, all public and invite-only review groups will be
        returned.

        For regular users, only review groups that are public or that the
        user is on the access list for will be returned.

        For anonymous users, only public review groups will be returned.

        The returned list is further filtered down based on the
        ``visible_only`` and ``local_site`` parameters.

        Note:
            This is not responsible for checking if a user has access to
            a given ``local_site``. This function assumes that access has
            already been checked.

        Version Changed:
            6.0:
            Removed the ``show_all_local_sites`` argument.

        Version Changed:
            5.0:
            Deprecated ``show_all_local_sites`` and added support for
            setting ``local_site`` to :py:class:`LocalSite.ALL
            <reviewboard.site.models.LocalSite.ALL>`.

        Version Changed:
            3.0.24:
            Added the ``distinct`` parameter.

        Args:
            user (django.contrib.auth.models.User):
                The user that must have access to any returned groups.

            visible_only (bool, optional):
                Whether only visible review groups should be returned.

            local_site (reviewboard.site.models.LocalSite or
                        reviewboard.site.models.LocalSite.ALL, optional):
                A specific :term:`Local Site` that the groups must be
                associated with. By default, this will only return groups
                not part of a site.

                This may be :py:attr:`LocalSite.ALL
                <reviewboard.site.models.LocalSite.ALL>`.

                Version Changed:
                    5.0:
                    Added support for :py:attr:`LocalSite.ALL
                    <reviewboard.site.models.LocalSite.ALL>`.

            distinct (bool, optional):
                Whether to return distinct results.

                Turning this off can increase performance. It's on by default
                for backwards-compatibility.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.
        """
        q = Q()

        if user.is_superuser:
            if visible_only:
                q &= Q(visible=True)
        else:
            if local_site is LocalSite.ALL:
                perm_local_site = None
            else:
                perm_local_site = local_site

            has_perm = user.has_perm('reviews.can_view_invite_only_groups',
                                     perm_local_site)

            if not has_perm:
                q &= Q(invite_only=False)

            if visible_only:
                # We allow accessible() to return hidden groups if the user is
                # a member, so we must perform this check here.
                q &= Q(visible=True)

            # Make sure we're only checking membership if we're checking
            # either visible= or invite_only= above, or users with the
            # special permission won't be able to look up repositories if
            # visible_only=False.
            if user.is_authenticated and (visible_only or not has_perm):
                q |= Q(users=user.pk)

        if local_site is not LocalSite.ALL:
            q &= Q(local_site=local_site)

        queryset = self.filter(q)

        if distinct:
            queryset = queryset.distinct()

        return queryset

    def accessible_ids(self, *args, **kwargs):
        """Return IDs of groups that are accessible by the given user.

        This wraps :py:meth:`accessible` and takes the same arguments
        (with the exception of ``distinct``, which is ignored).

        Version Changed:
            3.0.24:
            In prior versions, the order was not specified, but was
            generally numeric order. This should still be true, but
            officially, we no longer guarantee any order of results.

        Args:
            *args (tuple):
                Positional arguments to pass to :py:meth:`accessible`.

            **kwargs (dict):
                Keyword arguments to pass to :py:meth:`accessible`.

        Returns:
            list of int:
            The list of IDs.
        """
        kwargs['distinct'] = False

        return list(sorted(set(
            self.accessible(*args, **kwargs)
            .values_list('pk', flat=True)
        )))

    def can_create(self, user, local_site=None):
        """Returns whether the user can create groups."""
        return (user.is_superuser or
                (local_site and local_site.is_mutable_by(user)))


class ReviewRequestQuerySet(QuerySet):
    def with_counts(self, user):
        queryset = self

        if user and user.is_authenticated:
            select_dict = {}

            select_dict['new_review_count'] = """
                SELECT COUNT(*)
                  FROM reviews_review, accounts_reviewrequestvisit
                  WHERE reviews_review.public
                    AND reviews_review.review_request_id =
                        reviews_reviewrequest.id
                    AND accounts_reviewrequestvisit.review_request_id =
                        reviews_reviewrequest.id
                    AND accounts_reviewrequestvisit.user_id = %(user_id)s
                    AND reviews_review.timestamp >
                        accounts_reviewrequestvisit.timestamp
                    AND reviews_review.user_id != %(user_id)s
            """ % {
                'user_id': str(user.id)
            }

            queryset = self.extra(select=select_dict)

        return queryset


class ReviewRequestManager(ConcurrencyManager):
    """
    A manager for review requests. Provides specialized queries to retrieve
    review requests with specific targets or origins, and to create review
    requests based on certain data.
    """

    def get_queryset(self):
        """Return a QuerySet for ReviewRequest models.

        Returns:
            ReviewRequestQuerySet:
            The new QuerySet instance.
        """
        return ReviewRequestQuerySet(self.model)

    def create(self, user, repository, commit_id=None, local_site=None,
               create_from_commit_id=False, create_with_history=False):
        """Create a new review request.

        Args:
            user (django.contrib.auth.models.User):
                The user creating the review request. They will be tracked as
                the submitter.

            repository (reviewboard.scmtools.Models.Repository):
                The repository, if any, the review request is associated with.

                If ``None``, diffs cannot be added to the review request.

            commit_id (unicode, optional):
                An optional commit ID.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite to associate the review request with.

            create_from_commit_id (bool, optional):
                Whether or not the given ``commit_id`` should be used to
                pre-populate the review request data. If ``True``, the given
                ``repository`` will be used to do so.

            create_with_history (bool, optional):
                Whether or not the created review request will support
                attaching multiple commits per diff revision.

                If ``False``, it will not be possible to use the
                :py:class:`~reviewboard.webapi.resources.diff.DiffResource` to
                upload diffs; the
                :py:class:`~reviewboard.webapi.resources.DiffCommitResource`
                must be used instead.

        Returns:
            reviewboard.reviews.models.review_request.ReviewRequest:
            The created review request.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                The hosting service backing the repository encountered an
                error.

            reviewboard.scmtools.errors.ChangeNumberInUseError:
                The commit ID is already in use by another review request.

            reviewboard.scmtools.errors.SCMError:
                The repository tool encountered an error.

            ValueError:
                An invalid value was passed for an argument.
        """
        from reviewboard.reviews.models import ReviewRequestDraft

        if commit_id:
            # Try both the new commit_id and old changenum versions
            try:
                review_request = self.get(commit_id=commit_id,
                                          repository=repository)
                raise ChangeNumberInUseError(review_request)
            except ObjectDoesNotExist:
                pass

            try:
                draft = ReviewRequestDraft.objects.get(
                    commit_id=commit_id,
                    review_request__repository=repository)
                raise ChangeNumberInUseError(draft.review_request)
            except ObjectDoesNotExist:
                pass

            try:
                review_request = self.get(changenum=int(commit_id),
                                          repository=repository)
                raise ChangeNumberInUseError(review_request)
            except (ObjectDoesNotExist, TypeError, ValueError):
                pass

        if create_with_history:
            if repository is None:
                raise ValueError('create_with_history requires a repository.')
            elif create_from_commit_id:
                raise ValueError(
                    'create_from_commit_id and create_with_history cannot '
                    'both be set to True.')
            elif not repository.scmtool_class.supports_history:
                raise ValueError(
                    'This repository does not support review requests created '
                    'with history.')

        # Create the review request. We're not going to actually save this
        # until we're confident we have all the data we need.
        review_request = self.model(
            submitter=user,
            status='P',
            public=False,
            repository=repository,
            diffset_history=DiffSetHistory(),
            local_site=local_site)

        review_request.created_with_history = create_with_history

        if commit_id:
            review_request.commit = commit_id

        review_request.validate_unique()

        draft = None

        if commit_id and create_from_commit_id:
            try:
                draft = ReviewRequestDraft(review_request=review_request)
                draft.update_from_commit_id(commit_id)
            except Exception as e:
                logger.exception('Unable to update new review request from '
                                 'commit ID %s on repository ID=%s: %s',
                                 commit_id, repository.pk, e)
                raise

        # Now that we've guaranteed we have everything needed for this review
        # request, we can save all related objects and re-attach (since the
        # "None" IDs are cached).
        diffset_history = review_request.diffset_history
        diffset_history.save()
        review_request.diffset_history = diffset_history

        try:
            review_request.save()
        except IntegrityError as e:
            if 'changenum' in str(e):
                # We do have a race condition here where our check above may
                # have succeeded, but in the meantime another process ended up
                # creating the review request. This is more likely to happen
                # when the server is swamped and users start getting impatient.
                # In the case that we can't bail early, undo the objects we've
                # already created.
                if diffset_history:
                    diffset_history.delete()

                review_request = self.get(changenum=int(commit_id),
                                          repository=repository)
                raise ChangeNumberInUseError(review_request)
            else:
                raise

        if draft:
            draft.review_request = review_request
            draft.save()

            draft.add_default_reviewers()

            if draft.diffset and create_from_commit_id:
                # A diffset has been created from an existing commit. Now that
                # the review request draft, diffset and all of its related
                # objects have been created and saved to the database, we can
                # emit this signal.
                review_request_diffset_uploaded.send(
                    sender=self.__class__,
                    diffset=draft.diffset,
                    review_request_draft=draft)

        if local_site:
            # We want to atomically set the local_id to be a monotonically
            # increasing ID unique to the local_site. This isn't really
            # possible in django's DB layer, so we have to drop back to pure
            # SQL and then reload the model.
            from reviewboard.reviews.models import ReviewRequest

            with transaction.atomic():
                # TODO: Use the cursor as a context manager when we move over
                # to Django 1.7+.
                db = router.db_for_write(ReviewRequest)
                cursor = connections[db].cursor()
                cursor.execute(
                    'UPDATE %(table)s SET'
                    '  local_id = COALESCE('
                    '    (SELECT MAX(local_id) from'
                    '      (SELECT local_id FROM %(table)s'
                    '        WHERE local_site_id = %(local_site_id)s) as x'
                    '      ) + 1,'
                    '    1),'
                    '  local_site_id = %(local_site_id)s'
                    '    WHERE %(table)s.id = %(id)s' % {
                        'table': ReviewRequest._meta.db_table,
                        'local_site_id': local_site.pk,
                        'id': review_request.pk,
                    })
                cursor.close()

            review_request.local_id = (
                ReviewRequest.objects.filter(pk=review_request.pk)
                .values_list('local_id', flat=True)[0]
            )

        # Ensure that a draft exists, so that users will be prompted to publish
        # the new review request.
        ReviewRequestDraft.create(review_request)

        return review_request

    def get_to_group_query(self, group_name, local_site):
        """Return a Q() query object targeting a group.

        This is meant to be passed as an ``extra_query`` argument to
        :py:meth:`public`.

        Args:
            group_name (str):
                The name of the review group the review requests must be
                assigned to.

            local_site (reviewboard.site.models.LocalSite):
                The :term:`Local Site` that the review requests must be on,
                if any.

                This does not accept :py:attr:`LocalSite.ALL
                <reviewboard.site.models.LocalSite.ALL>`.

                Callers should first validate that the user has access to
                the Local Site, if provided.

        Returns:
            django.db.models.Q:
            The query object.
        """
        return (
            Q(target_groups__name=group_name) &
            LocalSite.objects.build_q(
                local_site,
                local_site_field='target_groups__local_site',
                allow_all=False)
        )

    def get_to_user_groups_query(self, user_or_username):
        """Return a Q() query object targeting groups joined by a user.

        This is meant to be passed as an ``extra_query`` argument to
        :py:meth:`public`.

        Args:
            user_or_username (django.contrib.auth.models.User or str):
                The User instance or username that all review requests must
                be assigned to indirectly.

        Returns:
            django.db.models.Q:
            The query object.

        Raises:
            django.contrib.auth.models.User.DoesNotExist:
                A username was provided, and that user does not exist.
        """
        query_user = self._get_query_user(user_or_username)
        groups = list(query_user.review_groups.values_list('pk', flat=True))

        return Q(target_groups__in=groups)

    def get_to_user_directly_query(self, user_or_username):
        """Returns the query targeting a user directly.

        This will include review requests where the user has been listed
        as a reviewer, or the user has starred.

        This is meant to be passed as an ``extra_query`` argument to
        :py:meth:`public`.

        Args:
            user_or_username (django.contrib.auth.models.User or str):
                The User instance or username that all review requests must
                be assigned to directly.

        Returns:
            django.db.models.Q:
            The query object.

        Raises:
            django.contrib.auth.models.User.DoesNotExist:
                A username was provided, and that user does not exist.
        """
        query_user = self._get_query_user(user_or_username)

        q = Q(Exists(
            self.model.target_people.through.objects
            .filter(
                Q(reviewrequest_id=OuterRef('pk')) &
                Q(user=query_user)
            )
        ))

        try:
            # Note that this should be a LEFT OUTER JOIN, so we'll be matching
            # the profiles but they shouldn't result in duplicates.
            profile = query_user.get_profile()
            q |= Q(starred_by=profile)
        except ObjectDoesNotExist:
            pass

        return q

    def get_to_user_query(self, user_or_username):
        """Return a Q() query object targeting a user indirectly.

        This will include review requests where the user has been listed
        as a reviewer, or a group that the user belongs to has been listed,
        or the user has starred.

        This is meant to be passed as an ``extra_query`` argument to
        :py:meth:`public`.

        Args:
            user_or_username (django.contrib.auth.models.User or str):
                The User instance or username that all review requests must
                be assigned to (directly to indirectly).

        Returns:
            django.db.models.Q:
            The query object.

        Raises:
            django.contrib.auth.models.User.DoesNotExist:
                A username was provided, and that user does not exist.
        """
        query_user = self._get_query_user(user_or_username)
        groups = list(query_user.review_groups.values_list('pk', flat=True))

        q = Q(Exists(
            self.model.target_people.through.objects
            .filter(
                Q(reviewrequest_id=OuterRef('pk')) &
                Q(user=query_user)
            )
        )) | Q(Exists(
            self.model.target_groups.through.objects
            .filter(
                Q(reviewrequest_id=OuterRef('pk')) &
                Q(group__in=groups)
            )
        ))

        try:
            profile = query_user.get_profile()
            q |= Q(starred_by=profile)
        except ObjectDoesNotExist:
            pass

        return q

    def get_from_user_query(self, user_or_username):
        """Return a Q() query object for review requests owned by a user.

        This is meant to be passed as an ``extra_query`` argument to
        :py:meth:`public`.

        Args:
            user_or_username (django.contrib.auth.models.User or str):
                The User instance or username that all review requests must
                be owned by.

        Returns:
            django.db.models.Q:
            The query object.
        """
        if isinstance(user_or_username, User):
            return Q(submitter=user_or_username)
        else:
            return Q(submitter__username=user_or_username)

    def get_to_or_from_user_query(self, user_or_username):
        """Return a Q() query object for review requests involving a user.

        This is meant to be passed as an ``extra_query`` argument to
        :py:meth:`public`.

        Args:
            user_or_username (django.contrib.auth.models.User or unicode):
                The User instance or username that all review requests must
                either be owned by or assigned to (directly to indirectly).

        Returns:
            django.db.models.Q:
            The query object.

        Raises:
            django.contrib.auth.models.User.DoesNotExist:
                A username was provided, and that user does not exist.
        """
        return (self.get_to_user_query(user_or_username) |
                self.get_from_user_query(user_or_username))

    def public(self, filter_private=True, *args, **kwargs):
        """Query public review requests, filtered by given criteria.

        Args:
            filter_private (bool, optional):
                Whether to filter out any review requests on private
                repositories or invite-only review groups that the user
                does not have access to.

                By default, they are filtered out.

                This requires ``user`` to be provided.

            *args (tuple):
                Additional positional arguments to pass to the common
                :py:meth:`_query` function.

            **kwargs (dict):
                Additional keyword arguments to pass to the common
                :py:meth:`_query` function.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.
        """
        return self._query(filter_private=filter_private, *args, **kwargs)

    def to_group(self, group_name, local_site, *args, **kwargs):
        """Query review requests made to a review group.

        The result will be review requests assigned to a review group.

        By default, the results will not be filtered based on whether a user
        has access to the review requests (via private repository or
        invite-only review group ACLs). To filter based on access, pass
        ``filter_private=True``.

        Args:
            group_name (str):
                The name of the review group the review requests must be
                assigned to.

            local_site (reviewboard.site.models.LocalSite):
                The :term:`Local Site` that the review requests must be on,
                if any.

                This does not accept :py:attr:`LocalSite.ALL
                <reviewboard.site.models.LocalSite.ALL>`.

                Callers should first validate that the user has access to
                the Local Site, if provided.

            *args (tuple):
                Additional positional arguments to pass to the common
                :py:meth:`_query` function.

            **kwargs (dict):
                Additional keyword arguments to pass to the common
                :py:meth:`_query` function.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.
        """
        return self._query(
            extra_query=self.get_to_group_query(group_name, local_site),
            local_site=local_site,
            *args, **kwargs)

    def to_or_from_user(self, user_or_username, *args, **kwargs):
        """Query review requests a user is involved in.

        The result will be review requests from a user, assigned to the user,
        or assigned to a group the user is in.

        By default, the results will not be filtered based on whether a user
        has access to the review requests (via private repository or
        invite-only review group ACLs). To filter based on access, pass
        ``filter_private=True``.

        Args:
            user_or_username (django.contrib.auth.models.User or unicode):
                The User instance or username that all review requests must
                either be owned by or assigned to (directly to indirectly).

            *args (tuple):
                Additional positional arguments to pass to the common
                :py:meth:`_query` function.

            **kwargs (dict):
                Additional keyword arguments to pass to the common
                :py:meth:`_query` function.

        Returns:
            django.db.models.query.QuerySet:
            A queryset of all review requests the users is involved in as
            either a submitter or a reviewer (either directly assigned or
            indirectly as a member of a group).

        Raises:
            django.contrib.auth.models.User.DoesNotExist:
                A username was provided, and that user does not exist.
        """
        return self._query(
            extra_query=self.get_to_or_from_user_query(user_or_username),
            *args, **kwargs)

    def to_user_groups(self, username, *args, **kwargs):
        """Query review requests made to a user's review groups.

        The result will be review requests assigned to a group the user is in.

        By default, the results will not be filtered based on whether a user
        has access to the review requests (via private repository or
        invite-only review group ACLs). To filter based on access, pass
        ``filter_private=True``.

        Args:
            username (django.contrib.auth.models.User or str):
                The User instance or username.

        Returns:
            django.db.models.query.QuerySet:
            A queryset of all review requests the users is involved in as
            either a submitter or a reviewer (either directly assigned or
            indirectly as a member of a group).

        Raises:
            django.contrib.auth.models.User.DoesNotExist:
                A username was provided, and that user does not exist.
        """
        return self._query(
            extra_query=self.get_to_user_groups_query(username),
            *args, **kwargs)

    def to_user_directly(self, user_or_username, *args, **kwargs):
        """Query review requests assigned directly to a user.

        The result will be review requests assigned to the user.

        By default, the results will not be filtered based on whether a user
        has access to the review requests (via private repository or
        invite-only review group ACLs). To filter based on access, pass
        ``filter_private=True``.

        Args:
            user_or_username (django.contrib.auth.models.User or unicode):
                The user object or username to query for.

            *args (tuple):
                Additional positional arguments to pass to the common
                :py:meth:`_query` function.

            **kwargs (dict):
                Additional keyword arguments to pass to the common
                :py:meth:`_query` function.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.

        Raises:
            django.contrib.auth.models.User.DoesNotExist:
                A username was provided, and that user does not exist.
        """
        return self._query(
            extra_query=self.get_to_user_directly_query(user_or_username),
            *args, **kwargs)

    def to_user(self, user_or_username, *args, **kwargs):
        """Query review requests assigned directly or indirectly to a user.

        The result will be review requests assigned to the user or to a group
        the user is in.

        By default, the results will not be filtered based on whether a user
        has access to the review requests (via private repository or
        invite-only review group ACLs). To filter based on access, pass
        ``filter_private=True``.

        Args:
            user_or_username (django.contrib.auth.models.User or unicode):
                The user object or username to query for.

            *args (tuple):
                Additional positional arguments to pass to the common
                :py:meth:`_query` function.

            **kwargs (dict):
                Additional keyword arguments to pass to the common
                :py:meth:`_query` function.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.

        Raises:
            django.contrib.auth.models.User.DoesNotExist:
                A username was provided, and that user does not exist.
        """
        return self._query(
            extra_query=self.get_to_user_query(user_or_username),
            *args, **kwargs)

    def from_user(self, user_or_username, *args, **kwargs):
        """Query review requests from a user.

        The result will be review requests created or currently owned by a
        user.

        By default, the results will not be filtered based on whether a user
        has access to the review requests (via private repository or
        invite-only review group ACLs). To filter based on access, pass
        ``filter_private=True``.

        Args:
            user_or_username (django.contrib.auth.models.User or unicode):
                The user object or username to query for.

            *args (tuple):
                Additional positional arguments to pass to the common
                :py:meth:`_query` function.

            **kwargs (dict):
                Additional keyword arguments to pass to the common
                :py:meth:`_query` function.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.
        """
        return self._query(
            extra_query=self.get_from_user_query(user_or_username),
            *args, **kwargs)

    def _query(
        self,
        user: Optional[User] = None,
        status: Optional[str] = 'P',
        with_counts: bool = False,
        extra_query: Optional[Q] = None,
        local_site: AnyOrAllLocalSites = None,
        filter_private: bool = False,
        show_inactive: bool = False,
        show_all_unpublished: bool = False,
        distinct: bool = False,
    ) -> QuerySet[ReviewRequest]:
        """Return a queryset for review requests matching the given criteria.

        By default, the results will not be filtered based on whether a user
        has access to the review requests (via private repository or
        invite-only review group ACLs). To filter based on access, pass
        ``filter_private=True``.

        Version Changed:
            6.0.2:
            * Added the ``distinct`` argument from Review Board 5.0.7.

        Version Changed:
            6.0:
            Removed the ``show_all_local_sites`` argument.

        Version Changed:
            5.0.7:
            * Added the ``distinct`` argument.

        Version Changed:
            5.0:
            Deprecated ``show_all_local_sites`` and added support for
            setting ``local_site`` to :py:class:`LocalSite.ALL
            <reviewboard.site.models.LocalSite.ALL>`.

        Args:
            user (django.contrib.auth.models.User, optional):
                The user that must have access to any returned review
                requests, if limiting access by user.

            status (str, optional):
                A review request status to filter by.

            with_counts (bool, optional):
                Whether to include new review counts since the last
                visit.

                If set, this will make use of
                :py:meth:`ReviewRequestQuerySet.with_counts`.

            extra_query (django.db.models.Q, optional):
                Optional additional queries for the queryset.

            local_site (reviewboard.site.models.LocalSite or
                        reviewboard.site.models.LocalSite.ALL, optional):
                A specific :term:`Local Site` that the groups must be
                associated with. By default, this will only return groups
                not part of a site.

                This may be :py:attr:`LocalSite.ALL
                <reviewboard.site.models.LocalSite.ALL>`.

                Callers should first validate that the user has access to
                the Local Site, if provided.

                Version Changed:
                    5.0:
                    Added support for :py:attr:`LocalSite.ALL
                    <reviewboard.site.models.LocalSite.ALL>`.

            filter_private (bool, optional):
                Whether to filter out any review requests on private
                repositories or invite-only review groups that the user
                does not have access to.

                This requires ``user`` to be provided.

            show_inactive (bool, optional):
                Whether to filter out review requests for inactive users.

            show_all_unpublished (bool, optional):
                Whether to include review requests not yet published.

            distinct (bool, optional):
                Whether to return distinct results.

                This is off by default, and rarely would need to be enabled.

                Version Added:
                    5.0.7, 6.0.2

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.
        """
        from reviewboard.reviews.models import Group

        is_authenticated = (user is not None and user.is_authenticated)

        if show_all_unpublished:
            q = Q()
        else:
            q = Q(public=True)

            if is_authenticated:
                q |= Q(submitter=user)

        if not show_inactive:
            q &= Q(submitter__is_active=True)

        if status:
            q &= Q(status=status)

        q &= LocalSite.objects.build_q(local_site)

        if extra_query:
            q &= extra_query

        if filter_private and (not user or not user.is_superuser):
            # This must always be kept in sync with RBSearchForm.search.
            model = self.model
            target_groups_m2m = model.target_groups.through.objects

            if is_authenticated:
                accessible_repo_ids = Repository.objects.accessible_ids(
                    user=user,
                    visible_only=False,
                    local_site=local_site)

                # This is not a subquery, and will operate directly on the
                # repository relation field on the review request.
                repo_q = (
                    Q(repository=None) |
                    Q(repository__in=accessible_repo_ids)
                )

                # This subquery will match if any target_people m2m entries
                # exist where the user is a target reviewer for the review
                # request. If there are no entries, this will not match.
                #
                # We use a subquery to avoid a JOIN here because we'd be
                # joining a one-to-many (review request to any through tables
                # matching it), which would lead to duplicates, and we don't
                # want to use DISTINCT.
                people_q = Q(Exists(
                    model.target_people.through.objects
                    .filter(
                        Q(reviewrequest_id=OuterRef('pk')) &
                        Q(user=user)
                    )
                ))

                # For the groups subquery, we need to match under any of the
                # following conditions:
                #
                # 1. There are no review groups.
                # 2. There are only accessible review groups.
                # 3. There are both accessible and inaccessible review groups.
                #
                # For this, we can't just check for the existence of
                # accessible review groups, because we'd fail case #1.
                #
                # We also can't exclude the existence of inaccessible review
                # groups (e.g., `~Exists(...~Q(group__in...))`), because
                # of case #3.
                #
                # So we have to factor in both pieces of information by using
                # two subqueries, we can do this by saying it's a match if
                # either of the following is true:
                #
                # 1. There are no review groups OR
                # 2. There's at least one accessible review group
                #
                # The SQL engine should short-circuit the second subquery if
                # the first says there are no review groups.
                #
                # Like above, we want to use a subquery for performance and
                # to avoid duplicates. The original (pre-5.0.7/6.0.2 code)
                # used two INNER JOINs (through table and reviews_group table)
                # and a DISTINCT to counteract those duplicates, which is slow.
                #
                # Since we're using Exists(), we stop on the first match of
                # either, so in the worst case, it should still be faster than
                # using those two INNER JOINs + DISTINCT.
                accessible_group_ids = Group.objects.accessible_ids(
                    user=user,
                    visible_only=False,
                    local_site=local_site)

                group_q = Q(
                    Q(~Exists(
                        target_groups_m2m
                        .filter(reviewrequest_id=OuterRef('pk'))
                    )) |
                    Q(Exists(
                        target_groups_m2m
                        .filter(
                            Q(reviewrequest_id=OuterRef('pk')) &
                            Q(group__in=accessible_group_ids)
                        )
                    ))
                )

                q &= (
                    Q(submitter=user) |
                    (repo_q &
                     (people_q |
                      group_q))
                )
            else:
                # This subquery will kick in for any non-NULL repositories,
                # and will check if the repository is public.
                repo_q = (
                    Q(repository=None) |
                    Q(Exists(
                        Repository.objects
                        .filter(
                            Q(pk=OuterRef('repository_id')) &
                            Q(public=True)
                        )
                    ))
                )

                # Same as above, we need to effectively perform two subqueries
                # in order to detect the "no review groups" case. Here, we
                # say it's a match if one of the following is true:
                #
                # 1. There are no review groups OR
                # 2. There's at least one public review group
                group_q = Q(
                    Q(~Exists(
                        target_groups_m2m
                        .filter(reviewrequest_id=OuterRef('pk'))
                    )) |
                    Q(Exists(
                        target_groups_m2m
                        .filter(
                            Q(reviewrequest_id=OuterRef('pk')) &
                            Q(group__invite_only=False)
                        )
                    ))
                )

                q &= repo_q & group_q

        queryset = self.filter(q)

        if distinct:
            queryset = queryset.distinct()

        if with_counts:
            queryset = queryset.with_counts(user)

        return queryset

    def _get_query_user(self, user_or_username):
        """Return a User object, given a possible User or username.

        If a User instance is provided, it will be directly returned.

        If a username is provided, it will be looked up and then returned.

        Args:
            user_or_username (django.contrib.auth.models.User or str):
                The User instance or username to look up.

        Returns:
            django.contrib.auth.models.User:
            The resulting User instance.

        Raises:
            django.contrib.auth.models.User.DoesNotExist:
                A username was provided, and that user does not exist.
        """
        if isinstance(user_or_username, User):
            return user_or_username
        else:
            return User.objects.get(username=user_or_username)

    def for_id(self, pk, local_site=None):
        """Returns the review request matching the given ID and LocalSite.

        If a LocalSite is provided, then the ID will be matched against the
        displayed ID for the LocalSite, rather than the in-database ID.
        """
        if local_site is None:
            return self.model.objects.get(pk=pk)
        else:
            return self.model.objects.get(Q(local_id=pk) &
                                          Q(local_site=local_site))


class _ReviewANY:
    """A sentinel indicating that reviews can be replying to any (or none).

    Usually we want to only query for top-level reviews, or we want all replies
    to a specific review. However, on some cases we want to allow queries to
    fetch all reviews, whether they're top-level or replies.

    Version Added:
        6.0
    """

    def __repr__(self) -> str:
        """Return a string representation of the sentinel.

        Returns:
            str:
            The string representation.
        """
        return '<ReviewManager.ANY>'


class ReviewManager(ConcurrencyManager):
    """A manager for Review models.

    This handles concurrency issues with Review models. In particular, it
    will try hard not to save two reviews at the same time, and if it does
    manage to do that (which may happen for pending reviews while a server
    is under heavy load), it will repair and consolidate the reviews on
    load. This prevents errors and lost data.
    """

    ANY = _ReviewANY()

    def accessible(
        self,
        user: User,
        extra_query: Optional[Q] = None,
        local_site: Optional[LocalSite] = None,
        public: Optional[bool] = None,
        base_reply_to: Optional[Union[Review, _ReviewANY]] = None,
        distinct: Optional[bool] = False,
    ) -> QuerySet:
        """Return a queryset for reviews accessible by the given user.

        For superusers, all public (published) and unpublished reviews will
        be returned.

        For regular users, only reviews that are owned by the user or that
        are public in the repositories, local sites, and review groups which
        the user has access to will be returned.

        For anonymous users, only public reviews that are on public
        repositories and whose review requests are not targeted by invite-only
        review groups will be returned.

        Version Changed:
            6.0:
            Added the ``base_reply_to`` argument.

        Version Added:
            5.0

        Args:
            user (django.contrib.auth.models.User):
                The User object that must have access to any returned reviews.

            extra_query (django.db.models.Q, optional):
                Additional query parameters to add for filtering
                down the resulting queryset.

            local_site (reviewboard.site.models.LocalSite or
                        reviewboard.site.models.LocalSite.ALL, optional):
                A specific :term:`Local Site` that the reviews must be
                associated with. It is assumed that the given user has access
                to the :term:`Local Site`. By default, this will only return
                reviews not part of a site.

                This may be :py:attr:`LocalSite.ALL
                <reviewboard.site.models.LocalSite.ALL>`.

            public (bool or None, optional):
                Whether to filter for public (published) reviews. If set to
                ``None``, both published and unpublished reviews will be
                included.

            base_reply_to (reviewboard.reviews.models.review.Review, optional):
                If provided, limit results to reviews that are part of the
                thread of replies to this review.

            distinct (bool, optional):
                Whether to return distinct results.

                Turning this on can decrease performance. It's off by default.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.
        """
        assert isinstance(user, (User, AnonymousUser))

        return self._query(user=user,
                           public=public,
                           extra_query=extra_query,
                           local_site=local_site,
                           status=None,
                           filter_private=True,
                           base_reply_to=base_reply_to,
                           distinct=distinct)

    def get_pending_review(self, review_request, user):
        """Return a user's pending review on a review request.

        This will handle fixing duplicate reviews if more than one pending
        review is found.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request being reviewed.

            user (django.contrib.auth.models.User):
                The user making the review.

        Returns:
            reviewboard.reviews.models.review.Review:
            The pending review object.
        """
        if not user.is_authenticated:
            return None

        reviews = list(
            self.filter(user=user,
                        review_request=review_request,
                        public=False,
                        base_reply_to__isnull=True)
            .order_by('timestamp')
        )

        if len(reviews) == 0:
            return None
        elif len(reviews) == 1:
            return reviews[0]
        else:
            # We have duplicate reviews, which will break things. We need
            # to condense them.
            logger.warning('Duplicate pending reviews found for review '
                           'request ID %s, user %s. Fixing.',
                           review_request.id, user.username)

            return self.fix_duplicate_reviews(reviews)

    def get_pending_reply(self, review, user):
        """Return a user's pending reply to a given review.

        This will handle fixing duplicate reviews if more than one pending
        review reply is found.

        Args:
            review (reviewboard.reviews.models.review.Review):
                The review being replied to.

            user (django.contrib.auth.models.User):
                The user making the reply.

        Returns:
            reviewboard.reviews.models.review.Review:
            The pending review object.
        """
        if not user.is_authenticated:
            return None

        reviews = list(
            self.filter(user=user,
                        public=False,
                        base_reply_to=review)
            .order_by('timestamp')
        )

        if len(reviews) == 0:
            return None
        elif len(reviews) == 1:
            return reviews[0]
        else:
            # We have duplicate replies, which will break things. We need
            # to condense them.
            logger.warning('Duplicate pending replies found for review '
                           'ID %s, user %s. Fixing.',
                           review.id, user.username)

            return self.fix_duplicate_reviews(reviews)

    def fix_duplicate_reviews(self, reviews):
        """Fix duplicate reviews, condensing them into a single review.

        This will consolidate the data from all reviews into the first
        review in the list, and return the first review.

        Args:
            reviews (list of reviewboard.reviews.models.review.Review):
                The list of duplicate reviews.

        Returns:
            reviewboard.reviews.models.review.Review:
            The first review in the list containing the consolidated data.
        """
        master_review = reviews[0]

        for review in reviews[1:]:
            for attname in ["body_top", "body_bottom", "body_top_reply_to",
                            "body_bottom_reply_to"]:
                review_value = getattr(review, attname)

                if (review_value and not getattr(master_review, attname)):
                    setattr(master_review, attname, review_value)

            for attname in ["comments", "screenshot_comments",
                            "file_attachment_comments",
                            "general_comments"]:
                master_m2m = getattr(master_review, attname)
                review_m2m = getattr(review, attname)

                for obj in review_m2m.all():
                    master_m2m.add(obj)
                    review_m2m.remove(obj)

            master_review.save()
            review.delete()

        return master_review

    def from_user(self, user_or_username, *args, **kwargs):
        """Return the query for reviews created by a user.

        Args:
            user_or_username (django.contrib.auth.models.User or str):
                The User object or username.

            *args (tuple):
                Additional positional arguments to pass to the common
                :py:meth:`_query` function.

            **kwargs (dict):
                Additional keyword arguments to pass to the common
                :py:meth:`_query` function.

        Returns:
            django.db.models.query.QuerySet:
            A queryset for all the reviews created by the given user.
        """
        if isinstance(user_or_username, User):
            extra_query = Q(user=user_or_username)
        else:
            assert isinstance(user_or_username, str)
            extra_query = Q(user__username=user_or_username)

        return self._query(extra_query=extra_query, *args, **kwargs)

    def _query(
        self,
        user: Optional[User] = None,
        public: Optional[bool] = None,
        status: Optional[str] = 'P',
        extra_query: Optional[Q] = None,
        local_site: Optional[LocalSite] = None,
        filter_private: Optional[bool] = False,
        base_reply_to: Union[Review, _ReviewANY, None] = None,
        distinct: Optional[bool] = False):
        """Do a query for reviews.

        Version Changed:
            5.0:
            Added the ``distinct`` parameter.

        Args:
            user (django.contrib.auth.models.User, optional):
                A user to query for.

            public (bool or None, optional):
                Whether to filter for public (published) reviews. If set to
                ``None``, both published and unpublished reviews will be
                included.

            status (unicode, optional):
                The status of the review request that reviews are associated
                with.

            extra_query (django.db.models.Q, optional):
                Additional query parameters to add.

            local_site (reviewboard.site.models.LocalSite or
                        reviewboard.site.models.LocalSite.ALL, optional):
                A local site to limit to, if appropriate. If a user is given,
                it is assumed that they have access to the :term:`Local Site`.
                By default, this will only return reviews not part of a site.

                This may be :py:attr:`LocalSite.ALL
                <reviewboard.site.models.LocalSite.ALL>`.


                Callers should first validate that the user has access to
                the Local Site, if provided.

            filter_private (bool, optional):
                Whether to filter out reviews from review requests on private
                repositories or invite-only review groups that the user
                does not have access to. This will also filter out unpublished
                reviews that are not owned by the user.

                This requires ``user`` to be provided.

            base_reply_to (reviewboard.reviews.models.review.Review, optional):
                If provided, limit results to reviews that are part of the
                thread of replies to this review.

            distinct (bool, optional):
                Whether to return distinct results.

                Turning this on can decrease performance. It's off by default.

        Returns:
            django.db.models.query.QuerySet:
            A queryset for the given conditions.
        """
        from reviewboard.reviews.models import Group

        q = Q()

        if base_reply_to is not ReviewManager.ANY:
            q &= Q(base_reply_to=base_reply_to)

        if status:
            q &= Q(review_request__status=status)

        q &= LocalSite.objects.build_q(
            local_site,
            local_site_field='review_request__local_site')

        if extra_query:
            q &= extra_query

        if filter_private and (not user or not user.is_superuser):
            repo_q = Q(review_request__repository=None)
            group_q = Q(review_request__target_groups=None)

            # TODO: should be consolidated with queries in ReviewRequestManager
            if user and user.is_authenticated:
                accessible_repo_ids = Repository.objects.accessible_ids(
                    user=user,
                    visible_only=False,
                    local_site=local_site)
                accessible_group_ids = Group.objects.accessible_ids(
                    user=user,
                    visible_only=False,
                    local_site=local_site)

                repo_q |= \
                    Q(review_request__repository__in=accessible_repo_ids)
                group_q |= \
                    Q(review_request__target_groups__in=accessible_group_ids)

                acl_check_q = (
                    repo_q &
                    (Q(review_request__target_people=user) |
                     group_q)
                )

                if public is None:
                    q &= (
                        Q(user=user) |
                        (Q(public=True) &
                         acl_check_q)
                    )
                elif public:
                    q &= Q(public=True) & acl_check_q
                else:
                    q &= Q(public=False) & Q(user=user)
            else:
                # Return an empty result when an unauthenticated user queries
                # for unpublished reviews.
                if public is False:
                    return self.none()

                repo_q |= Q(review_request__repository__public=True)
                group_q |= Q(review_request__target_groups__invite_only=False)

                q &= repo_q & group_q & Q(public=True)
        else:
            if public is not None:
                q &= Q(public=public)

        queryset = self.filter(q)

        if distinct:
            queryset = queryset.distinct()

        return queryset


class StatusUpdateManager(Manager):
    """A manager for StatusUpdate models.

    This offers conveniences around creating
    :py:class:`~reviewboard.reviews.models.status_update.StatusUpdate` models
    for custom integrations.

    Version Added:
        5.0.3
    """

    def create_for_integration(
        self,
        integration: Integration,
        *,
        config: IntegrationConfig,
        user: User,
        review_request: ReviewRequest,
        change_description: Optional[ChangeDescription] = None,
        service_id: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        state: Optional[str] = None,
        can_retry: bool = False,
        extra_data: Dict = {},
        starting_description: str = 'starting...',
        waiting_description: str = 'waiting to run.',
        **kwargs,
    ) -> StatusUpdate:
        """Return a new status update for a given integration.

        This helps with generating defaults for a status update, and putting
        it in the correct initial state when running manually.

        The integration configuration will be associated with the status
        update, which is important for manually running integrations when
        multiple integrations are present on a review request.

        Args:
            integration (reviewboard.integrations.base.Integration):
                The integration that this status update will be associated
                with.

            config (reviewboard.integrations.models.IntegrationConfig):
                The configuration for the integration, used to provide
                defaults and used for later manual runs.

            user (django.contrib.auth.models.User):
                The user that the status update will be associated with.

            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request that the status update will be associated
                with.

            change_description (reviewboard.changedescs.models.
                                ChangeDescription, optional):
                The optional change description that the status update will
                be associated with.

            service_id (str, optional):
                An explicit service ID for the status update.

                If not provided (or if ``None``), a slugified version of
                the integration's name will be used.

            summary (str, optional):
                An explicit summary for the status update.

                If not provided (or if ``None``), the integration name will
                be used.

            description (str, optional):
                An explicit description for the status update.

                If not provided (or if ``None``), a standardized description
                will be used depending on whether the status update will be
                created in manual run mode.

                See ``starting_description`` and ``waiting_description`` to
                customize these strings.

            state (str, optional):
                An explicit state for the status update.

                If not provided (or if ``None``), the state will be in Not
                Yet Run if creating in manual run mode, or Pending otherwise.

            can_retry (bool, optional):
                Whether the status update can be retried after being run.

            extra_data (dict, optional):
                Extra data to store in the status update.

            starting_description (str, optional):
                The description to use if creating a status update that is
                immediately starting.

            waiting_description (str, optional):
                The description to use if creating a status update that is
                in manual run mode.

            **kwargs (dict, optional):
                Additional keyword arguments for the model.

        Returns:
            reviewboard.reviews.status_update.StatusUpdate:
            The new status update.
        """
        run_manually = config.get('run_manually')

        if description is None:
            if run_manually:
                description = waiting_description
            else:
                description = starting_description

        if state is None:
            if run_manually:
                state = self.model.NOT_YET_RUN
            else:
                state = self.model.PENDING

        if not service_id:
            assert integration.name
            service_id = slugify(integration.name)

        status_update: StatusUpdate = self.model(
            user=user,
            review_request=review_request,
            change_description=change_description,
            service_id=service_id,
            summary=summary or integration.name,
            description=description,
            state=state,
            extra_data=dict(extra_data, **{
                'can_retry': can_retry,
            }),
            **kwargs)
        status_update.integration_config = config
        status_update.save()

        return status_update



"""Unit tests for reviewboard.admin.siteconfig."""

import os

from django.conf import settings
from django.urls import reverse
from djblets.siteconfig.models import SiteConfiguration

from reviewboard.admin.siteconfig import load_site_config
from reviewboard.testing.testcase import TestCase


class LoadSiteConfigTests(TestCase):
    """Unit tests for reviewboard.admin.siteconfig.load_site_config."""

    def setUp(self):
        super(LoadSiteConfigTests, self).setUp()

        self.siteconfig = SiteConfiguration.objects.get_current()

    def test_with_site_domain_method_http(self):
        """Testing load_site_config with site_domain_method=http"""
        self.siteconfig.set('site_domain_method', 'http')
        self.siteconfig.save()
        load_site_config()

        self.assertEqual(str(os.environ.get(str('HTTPS'))), str('off'))
        self.assertFalse(getattr(settings, 'CSRF_COOKIE_SECURE', None))

        # Ensure that CSRF cookie flags are set correctly.
        self.create_user(username='test-user',
                         password='test-user')

        login_url = reverse('login')
        response = self.client.get(login_url)
        csrf_cookie = response.cookies.get(settings.CSRF_COOKIE_NAME)
        self.assertIsNotNone(csrf_cookie)
        self.assertFalse(csrf_cookie['secure'])

        # Ensure that session cookie flags are set correctly.
        response = self.client.post(
            reverse('login'),
            {
                'username': 'test-user',
                'password': 'test-user',
            })

        session_cookie = response.cookies.get(settings.SESSION_COOKIE_NAME)
        self.assertIsNotNone(session_cookie)
        self.assertFalse(session_cookie['secure'])

    def test_with_site_domain_method_https(self):
        """Testing load_site_config with site_domain_method=https"""
        self.siteconfig.set('site_domain_method', 'https')
        self.siteconfig.save()
        load_site_config()

        self.assertEqual(str(os.environ.get(str('HTTPS'))), str('on'))
        self.assertTrue(getattr(settings, 'CSRF_COOKIE_SECURE', None))

        # Ensure that CSRF cookie flags are set correctly.
        login_url = reverse('login')

        self.create_user(username='test-user',
                         password='test-user')

        response = self.client.get(login_url, secure=True)
        csrf_cookie = response.cookies.get(settings.CSRF_COOKIE_NAME)
        self.assertIsNotNone(csrf_cookie)
        self.assertTrue(csrf_cookie['secure'])

        # Ensure that session cookie flags are set correctly.
        response = self.client.post(
            login_url,
            {
                'username': 'test-user',
                'password': 'test-user',
            },
            secure=True)

        session_cookie = response.cookies.get(settings.SESSION_COOKIE_NAME)
        self.assertIsNotNone(session_cookie)
        self.assertTrue(session_cookie['secure'])



"""Unit tests for reviewboard.search.search_backends.base.SearchBackend."""

import kgb
from django.core.exceptions import ValidationError
from djblets.siteconfig.models import SiteConfiguration

from haystack.backends.simple_backend import SimpleSearchBackend
from reviewboard.search.search_backends.base import SearchBackend
from reviewboard.testing.testcase import TestCase


class TestSimpleSearchBackend(SearchBackend):
    search_backend_id = 'simple'
    name = 'Simple Engine'
    haystack_backend_name = \
        'haystack.backends.simple_backend.SimpleEngine'
    default_settings = {
        'setting1': 'value1',
        'setting2': 42,
        'setting3': [1, 2, 3],
    }


class SearchBackend(kgb.SpyAgency, TestCase):
    """Unit tests for reviewboard.search.search_backends.base.SearchBackend."""

    def setUp(self):
        super(SearchBackend, self).setUp()

        self.backend = TestSimpleSearchBackend()

    def test_configuration_getter_with_settings(self):
        """Testing SearchBackend.configuration getter with existing settings"""
        siteconfig = SiteConfiguration.objects.get_current()
        siteconfig.set('search_backend_settings', {
            'simple': {
                'setting1': 'new value',
                'setting3': [4, 5, 6],
            },
        })
        siteconfig.save(update_fields=('settings',))

        self.assertEqual(
            self.backend.configuration,
            {
                'ENGINE': 'haystack.backends.simple_backend.SimpleEngine',
                'setting1': 'new value',
                'setting2': 42,
                'setting3': [4, 5, 6],
            })

    def test_configuration_getter_without_settings(self):
        """Testing SearchBackend.configuration getter without any existing
        settings
        """
        siteconfig = SiteConfiguration.objects.get_current()
        siteconfig.set('search_backend_settings', {})
        siteconfig.save(update_fields=('settings',))

        self.assertEqual(
            self.backend.configuration,
            {
                'ENGINE': 'haystack.backends.simple_backend.SimpleEngine',
                'setting1': 'value1',
                'setting2': 42,
                'setting3': [1, 2, 3],
            })

        self.assertEqual(siteconfig.get('search_backend_settings'), {})

    def test_configuration_setter_with_existing_settings(self):
        """Testing SearchBackend.configuration setter with existing settings"""
        siteconfig = SiteConfiguration.objects.get_current()
        siteconfig.set('search_backend_settings', {
            'simple': {
                'setting1': 'new value',
                'setting3': [4, 5, 6],
            },
        })
        siteconfig.save(update_fields=('settings',))

        self.backend.configuration = {
            'setting2': 100,
            'setting4': 'new',
        }

        # Explicitly fetch a new instance, rather than reloading the cached
        # version.
        siteconfig = SiteConfiguration.objects.get(pk=siteconfig.pk)

        self.assertEqual(
            siteconfig.get('search_backend_settings'),
            {
                'simple': {
                    'setting1': 'new value',
                    'setting2': 100,
                    'setting3': [4, 5, 6],
                },
            })

    def test_configuration_setter_without_existing_settings(self):
        """Testing SearchBackend.configuration setter without existing
        settings
        """
        siteconfig = SiteConfiguration.objects.get_current()
        siteconfig.set('search_backend_settings', {})
        siteconfig.save(update_fields=('settings',))

        self.backend.configuration = {
            'setting2': 100,
            'setting4': 'new',
        }

        # Explicitly fetch a new instance, rather than reloading the cached
        # version.
        siteconfig = SiteConfiguration.objects.get(pk=siteconfig.pk)

        self.assertEqual(
            siteconfig.get('search_backend_settings'),
            {
                'simple': {
                    'setting2': 100,
                },
            })

    def test_validate(self):
        """Testing SearchBackend.validate"""
        self.backend.validate(configuration={
            'setting1': 'new value',
            'setting3': [4, 5, 6],
        })

    def test_validate_with_failure(self):
        """Testing SearchBackend.validate with failure"""
        self.spy_on(SimpleSearchBackend.search,
                    owner=SimpleSearchBackend,
                    func_name='search',
                    op=kgb.SpyOpRaise(Exception('Things went broken.')))

        message = (
            'Performing a test query failed. Make sure your configuration is '
            'correct. The error we received from the search backend was: '
            'Things went broken.'
        )

        with self.assertRaisesMessage(ValidationError, message):
            self.backend.validate(configuration={
                'setting1': 'new value',
                'setting3': [4, 5, 6],
            })



"""Unit tests for reviewboard.reviews.search_indexes.

Version Added:
    8.0
"""

from __future__ import annotations

import itertools
from typing import TYPE_CHECKING

from django.contrib.auth.models import User
from django.db.models import Q
from django_assert_queries.testing import assert_queries

from reviewboard.accounts.models import Profile
from reviewboard.diffviewer.models import DiffSet, DiffSetHistory, FileDiff
from reviewboard.reviews.models import Group, ReviewRequest
from reviewboard.reviews.search_indexes import ReviewRequestIndex
from reviewboard.scmtools.models import Repository
from reviewboard.search.testing import search_enabled
from reviewboard.site.models import LocalSite
from reviewboard.testing import TestCase

if TYPE_CHECKING:
    from django_assert_queries import ExpectedQueries


class ReviewRequestIndexTests(TestCase):
    """Unit tests for ReviewRequestIndex.

    Version Added:
        8.0
    """

    fixtures = ['test_users', 'test_scmtools']

    def test_update(self) -> None:
        """Testing ReviewRequestIndex.update"""
        # Set up a bunch of default objects.
        group1 = self.create_review_group(name='group1')
        group2 = self.create_review_group(name='group2',
                                          invite_only=True)

        user1 = self.create_user(username='user1')
        user2 = self.create_user(username='user2')
        user3 = self.create_user(username='user3')

        Profile.objects.create(user=user1)
        Profile.objects.create(user=user2,
                               is_private=True)

        repository1 = self.create_repository(name='repo1')
        repository2 = self.create_repository(name='repo2',
                                             public=False)
        repository3 = self.create_repository(name='repo3',
                                             public=False)
        repository3.users.add(user1, user3)

        target_people = [user1, user2, user3]

        diffset_histories: list[DiffSetHistory] = []
        review_requests: list[ReviewRequest] = []
        diffsets: list[DiffSet] = []

        for (user, repository, target_groups) in itertools.product(
            (user1, user2, user3),
            (repository1, repository2, None),
            ([group1], [group2]),
        ):
            review_request = self.create_review_request(
                publish=True,
                repository=repository,
                submitter=user,
                target_groups=target_groups,
                target_people=target_people,
            )
            review_requests.append(review_request)

            assert review_request.diffset_history is not None
            diffset_histories.append(review_request.diffset_history)

            if repository is not None:
                diffsets += [
                    self.create_diffset(review_request),
                    self.create_diffset(review_request),
                ]

        # Pre-populate caches.
        self.assertFalse(LocalSite.objects.has_local_sites())

        equeries: ExpectedQueries = [
            {
                'model': ReviewRequest,
                'only_fields': {
                    # ReviewRequest fields.
                    'bugs_closed',
                    'changenum',
                    'commit_id',
                    'description',
                    'last_updated',
                    'local_id',
                    'local_site_id',
                    'public',
                    'repository_id',
                    'submitter_id',
                    'summary',
                    'testing_done',

                    # DiffSetHistory fields.
                    'diffset_history__id',
                },
                'select_related': {
                    'diffset_history',
                },
                'where': Q(public=True)
            },
            {
                'model': Repository,
                'only_fields': {
                    'local_site',
                    'public',
                },
                'where': Q(id__in={
                    repository1.pk,
                    repository2.pk,
                    None,
                }),
            },
            {
                'model': User,
                'only_fields': {
                    # User fields.
                    'first_name',
                    'last_name',
                    'username',

                    # Profile fields.
                    'profile__extra_data',
                    'profile__is_private',
                },
                'select_related': {
                    'profile',
                },
                'where': Q(id__in={
                    user1.pk,
                    user2.pk,
                    user3.pk,
                }),
            },
            {
                'model': DiffSet,
                'only_fields': {
                    'history_id',
                },
                'where': Q(history__in=diffset_histories[::-1]),
            },
            {
                'model': FileDiff,
                'only_fields': {
                    'dest_file',
                    'source_file',
                },
                'where': Q(diffset__in=diffsets),
            },
            {
                'extra': {
                    '_prefetch_related_val_reviewrequest_id': (
                        '"reviews_reviewrequest_target_groups".'
                        '"reviewrequest_id"',
                        [],
                    ),
                },
                'join_types': {
                    'reviews_reviewrequest_target_groups': 'INNER JOIN',
                },
                'model': Group,
                'num_joins': 1,
                'only_fields': {
                    # Group fields.
                    'invite_only',
                    'local_site',
                },
                'tables': {
                    'reviews_group',
                    'reviews_reviewrequest_target_groups',
                },
                'where': Q(review_requests__in=review_requests[::-1]),
            },
            {
                'extra': {
                    '_prefetch_related_val_reviewrequest_id': (
                        '"reviews_reviewrequest_target_people".'
                        '"reviewrequest_id"',
                        [],
                    ),
                },
                'join_types': {
                    'reviews_reviewrequest_target_people': 'INNER JOIN',
                },
                'model': User,
                'num_joins': 1,
                'only_fields': {'id'},
                'tables': {
                    'auth_user',
                    'reviews_reviewrequest_target_people',
                },
                'where':
                    Q(directed_review_requests__in=review_requests[::-1]),
            },
        ]

        with search_enabled():
            with assert_queries(equeries):
                ReviewRequestIndex().update()

    def test_update_with_local_sites(self) -> None:
        """Testing ReviewRequestIndex.update with LocalSites"""
        # Set up a bunch of default objects.
        user1 = self.create_user(username='user1')
        user2 = self.create_user(username='user2')
        user3 = self.create_user(username='user3')

        Profile.objects.create(user=user1)
        Profile.objects.create(
            user=user2,
            is_private=True,
        )

        local_site1 = self.create_local_site(
            name='local-site-1',
            users=[user1],
        )

        local_site2 = self.create_local_site(
            name='local-site-2',
            users=[user2],
        )

        diffset_histories: list[DiffSetHistory] = []
        diffsets: list[DiffSet] = []
        repository_ids: set[int | None] = {None}
        review_requests: list[ReviewRequest] = []

        for local_site in (local_site1, local_site2, None):
            group1 = self.create_review_group(
                name='group1',
                local_site=local_site,
                users=[user1],
            )
            group2 = self.create_review_group(
                name='group2',
                local_site=local_site,
                invite_only=True,
                users=[user2],
            )

            repository1 = self.create_repository(
                name='repo1',
                local_site=local_site,
            )

            repository2 = self.create_repository(
                name='repo2',
                local_site=local_site,
                public=False,
            )
            repository2.review_groups.add(group2)

            repository3 = self.create_repository(
                name='repo3',
                local_site=local_site,
                public=False,
            )
            repository3.users.add(user3)

            target_people = [user1, user2, user3]

            local_id = 1

            for (user, repository, target_groups) in itertools.product(
                (user1, user2, user3),
                (repository1, repository2, None),
                ([group1], [group2]),
            ):
                review_request = self.create_review_request(
                    public=True,
                    local_id=local_id,
                    local_site=local_site,
                    repository=repository,
                    submitter=user,
                    target_groups=target_groups,
                    target_people=target_people,
                )
                review_requests.append(review_request)
                local_id += 1

                assert review_request.diffset_history is not None
                diffset_histories.append(review_request.diffset_history)

                if repository is not None:
                    repository_ids.add(repository.pk)
                    diffsets += [
                        self.create_diffset(review_request),
                        self.create_diffset(review_request),
                    ]

        # Pre-populate caches.
        self.assertTrue(LocalSite.objects.has_local_sites())

        equeries: ExpectedQueries = [
            {
                'model': ReviewRequest,
                'only_fields': {
                    # ReviewRequest fields.
                    'bugs_closed',
                    'changenum',
                    'commit_id',
                    'description',
                    'last_updated',
                    'local_id',
                    'local_site_id',
                    'public',
                    'repository_id',
                    'submitter_id',
                    'summary',
                    'testing_done',

                    # DiffSetHistory fields.
                    'diffset_history__id',
                },
                'select_related': {
                    'diffset_history',
                },
                'where': Q(public=True)
            },
            {
                'model': Repository,
                'only_fields': {
                    # Repository fields.
                    'local_site',
                    'public',

                    # LocalSite fields.
                    'local_site__public',
                },
                'select_related': {
                    'local_site',
                },
                'where': Q(id__in=repository_ids),
            },
            {
                'model': User,
                'only_fields': {
                    # User fields.
                    'first_name',
                    'last_name',
                    'username',

                    # Profile fields.
                    'profile__extra_data',
                    'profile__is_private',
                },
                'select_related': {'profile'},
                'where': Q(id__in={
                    user1.pk,
                    user2.pk,
                    user3.pk,
                }),
            },
            {
                'model': DiffSet,
                'only_fields': {
                    'history_id',
                },
                'where': Q(history__in=diffset_histories[::-1]),
            },
            {
                'model': FileDiff,
                'only_fields': {
                    'dest_file',
                    'source_file',
                },
                'where': Q(diffset__in=diffsets),
            },
            {
                'extra': {
                    '_prefetch_related_val_reviewrequest_id': (
                        '"reviews_reviewrequest_target_groups".'
                        '"reviewrequest_id"',
                        [],
                    ),
                },
                'join_types': {
                    'reviews_reviewrequest_target_groups': 'INNER JOIN',
                },
                'model': Group,
                'num_joins': 1,
                'only_fields': {
                    # Group fields.
                    'invite_only',
                    'local_site',

                    # LocalSite fields.
                    'local_site__public',
                },
                'select_related': {
                    'local_site',
                },
                'tables': {
                    'reviews_group',
                    'reviews_reviewrequest_target_groups',
                },
                'where': Q(review_requests__in=review_requests[::-1]),
            },
            {
                'extra': {
                    '_prefetch_related_val_reviewrequest_id': (
                        '"reviews_reviewrequest_target_people".'
                        '"reviewrequest_id"',
                        [],
                    ),
                },
                'join_types': {
                    'reviews_reviewrequest_target_people': 'INNER JOIN',
                },
                'model': User,
                'num_joins': 1,
                'only_fields': {'id'},
                'tables': {
                    'auth_user',
                    'reviews_reviewrequest_target_people',
                },
                'where':
                    Q(directed_review_requests__in=review_requests[::-1]),
            },
            {
                'model': LocalSite,
                'only_fields': {
                    'name',
                },
                'where': Q(id__in={
                    local_site1.pk,
                    local_site2.pk,
                    None,
                }),
            },
        ]

        with search_enabled():
            with assert_queries(equeries, with_tracebacks=True):
                ReviewRequestIndex().update()



"""Tests for reviewboard.extensions.hooks.ReviewRequestConditionChoicesHook.

Version Added:
    8.0
"""

from __future__ import annotations

from djblets.conditions.choices import (BaseConditionIntegerChoice,
                                        BaseConditionStringChoice)
from djblets.conditions.operators import (ConditionOperators,
                                          IsNotOneOfOperator,
                                          IsOneOfOperator,
                                          UnsetOperator)
from djblets.conditions.values import ConditionValueMultipleChoiceField
from reviewboard.extensions.hooks import ReviewRequestConditionChoicesHook
from reviewboard.reviews.conditions import (ReviewRequestConditionChoiceMixin,
                                            review_request_condition_choices)
from reviewboard.extensions.tests.testcases import BaseExtensionHookTestCase


class _MyCategoryChoice(ReviewRequestConditionChoiceMixin,
                        BaseConditionStringChoice):
    choice_id = 'my-category'
    name = 'Category'

    operators = ConditionOperators([
        UnsetOperator,
        IsOneOfOperator,
        IsNotOneOfOperator,
    ])

    default_value_field = ConditionValueMultipleChoiceField[str](choices=[
       ('architecture', 'Architecture'),
       ('bug', 'Bug'),
       ('docs', 'Documentation'),
       ('feature', 'Feature'),
       ('security', 'Security'),
    ])

    def get_match_value(self, review_request, **kwargs):
        return review_request.extra_data.get('my_category')


class _MyTaskIDChoice(ReviewRequestConditionChoiceMixin,
                      BaseConditionIntegerChoice):
    choice_id = 'my-task-id'
    name = 'Task ID'

    def get_match_value(self, review_request, **kwargs):
        # This would return an integer.
        return review_request.extra_data.get('my_task_id')


class ReviewRequestConditionChoicesHookTests(BaseExtensionHookTestCase):
    """Tests for ReviewRequestConditionChoicesHook."""

    def test_register(self) -> None:
        """Testing ReviewRequestConditionChoicesHook initializing"""
        ReviewRequestConditionChoicesHook(self.extension, [
            _MyCategoryChoice,
            _MyTaskIDChoice,
        ])

        self.assertIn(_MyCategoryChoice, review_request_condition_choices)
        self.assertIn(_MyTaskIDChoice, review_request_condition_choices)

    def test_unregister(self) -> None:
        """Testing ReviewRequestConditionChoicesHook uninitializing"""
        hook = ReviewRequestConditionChoicesHook(self.extension, [
            _MyCategoryChoice,
            _MyTaskIDChoice,
        ])
        hook.disable_hook()

        self.assertNotIn(_MyCategoryChoice, review_request_condition_choices)
        self.assertNotIn(_MyTaskIDChoice, review_request_condition_choices)



"""Unit tests for reviewboard.actions.renderers.DefaultActionGroupRenderer.

Version Added:
    8.0
"""

from __future__ import annotations

from django.template import Context
from django.utils.safestring import SafeString

from reviewboard.actions.renderers import DefaultActionGroupRenderer
from reviewboard.actions.tests.base import (TestActionsRegistry,
                                            TestGroupAction,
                                            TestGroupActionWithSubgroups,
                                            TestGroupItemAction1,
                                            TestGroupItemAction2,
                                            TestSubgroupAction)
from reviewboard.deprecation import RemovedInReviewBoard10_0Warning
from reviewboard.testing import TestCase


class DefaultActionGroupRendererTests(TestCase):
    """Unit tests for DefaultActionGroupRenderer.

    Version Added:
        8.0
    """

    def test_render(self) -> None:
        """Testing DefaultActionGroupRenderer.render"""
        action = TestGroupAction()
        placement = action.get_placement('review-request')

        item_action_1 = TestGroupItemAction1()
        item_action_2 = TestGroupItemAction2()

        registry = TestActionsRegistry()
        registry.register(action)
        registry.register(item_action_1)
        registry.register(item_action_2)

        renderer = DefaultActionGroupRenderer(action=action,
                                              placement=placement)
        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        html = renderer.render(request=request,
                               context=context)

        self.assertIsInstance(html, SafeString)
        self.assertHTMLEqual(
            html,
            """
            <li class="rb-c-actions__action"
                role="group">
             <a id="action-review-request-group-item-1-action"
                href="#"
                role="button">
              Group Item 1
             </a>
             <a id="action-review-request-group-item-2-action"
                href="#"
                role="button">
              Group Item 2
             </a>
            </li>
            """)

    def test_render_with_subgroups(self) -> None:
        """Testing DefaultActionGroupRenderer.render with subgroups"""
        # This renderer does not support subgroups.
        action = TestGroupActionWithSubgroups()
        placement = action.get_placement('header')

        item_action_1 = TestGroupItemAction1()
        item_action_2 = TestGroupItemAction2()
        subgroup = TestSubgroupAction()

        registry = TestActionsRegistry()
        registry.register(action)
        registry.register(subgroup)
        registry.register(item_action_1)
        registry.register(item_action_2)

        renderer = DefaultActionGroupRenderer(action=action,
                                              placement=placement)
        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        with self.assertLogs() as logs:
            html = renderer.render(request=request,
                                   context=context)

        self.assertEqual(
            logs.output,
            [
                "ERROR:reviewboard.actions.renderers:Could not render "
                "action 'subgroup-action' inside of group action "
                "'group-with-subgroups-action' in attachment point "
                "'header'. This location does not allow for nesting of "
                "groups.",
            ])

        self.assertIsInstance(html, SafeString)
        self.assertHTMLEqual(
            html,
            """
            <li class="rb-c-actions__action"
                role="group">
             <a id="action-header-group-item-1-action"
                href="#"
                role="button">
              Group Item 1
             </a>
            </li>
            """)

    def test_render_with_action_template(self) -> None:
        """Testing DefaultActionGroupRenderer.render with action.template_name
        set
        """
        class MyAction(TestGroupAction):
            template_name = 'actions/button_action.html'

        # We already test for the deprecation warning message in the action
        # tests. This just suppresses warning output for the test run.
        with self.assertWarns(RemovedInReviewBoard10_0Warning):
            action = MyAction()

        placement = action.get_placement('review-request')

        registry = TestActionsRegistry()
        registry.register(action)

        renderer = DefaultActionGroupRenderer(action=action,
                                              placement=placement)
        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        html = renderer.render(request=request,
                               context=context)

        self.assertIsInstance(html, SafeString)
        self.assertHTMLEqual(
            html,
            """
            <li class="rb-c-actions__action"
                role="presentation">
             <button aria-label="Test Group"
                     class="ink-c-button"
                     id="action-review-request-group-action"
                     type="button">
              <label class="ink-c-button__label">
               Test Group
              </label>
             </button>
            </li>
            """)

    def test_render_js(self) -> None:
        """Testing DefaultActionGroupRenderer.render_js"""
        action = TestGroupAction()
        placement = action.get_placement('review-request')

        registry = TestActionsRegistry()
        registry.register(action)

        renderer = DefaultActionGroupRenderer(action=action,
                                              placement=placement)
        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        js = renderer.render_js(request=request,
                                context=context)

        self.assertIsInstance(js, SafeString)
        self.assertHTMLEqual(
            js,
            """
            page.addActionView(new RB.Actions.ActionView({
                "attachmentPointID": "review-request",
                el: $('#action-review-request-group-action'),
                model: page.getAction("group-action"),
            }));
            """)



"""Management command to manually update the state of an update check."""

from django.core.management.base import BaseCommand, CommandError
from django.utils.translation import gettext as _
from djblets.siteconfig.models import SiteConfiguration


class Command(BaseCommand):
    """Management command to manually update the state of an update check."""

    help = _('Resolves a manual update check.')

    def add_arguments(self, parser):
        """Add arguments to the command.

        Args:
            parser (argparse.ArgumentParser):
                The argument parser for the command.
        """
        parser.add_argument(
            'check_name',
            metavar='NAME',
            help=_('The name of the check to resolve.'))

    def handle(self, **options):
        """Handle the command.

        Args:
            **options (dict, unused):
                Options parsed on the command line.

        Raises:
            django.core.management.CommandError:
                There was an error with arguments.
        """
        check_name = options['check_name']

        siteconfig = SiteConfiguration.objects.get_current()
        updates = siteconfig.settings.get('manual-updates', {})

        if check_name not in updates:
            raise CommandError(_('Couldn\'t find manual update check "%s"')
                               % check_name)

        if updates[check_name]:
            self.stdout.write(_('Already resolved manual update check "%s"')
                              % check_name)
        else:
            updates[check_name] = True
            siteconfig.save()

            self.stdout.write(_('Resolved manual update check "%s"')
                              % check_name)



"""Unit tests for reviewboard.site.templatetags.local_site."""

from django.template import Context, Template

from reviewboard.testing.testcase import TestCase


class TemplateTagTests(TestCase):
    """Unit tests for reviewboard.site.templatetags.local_site."""

    def test_local_site_url_with_no_local_site(self):
        """Testing localsite's {% url %} with no local site"""
        context = Context({})

        t = Template('{% url "dashboard" %}')
        self.assertEqual(t.render(context), '/dashboard/')

        t = Template('{% url "user" "sample-user" %}')
        self.assertEqual(t.render(context), '/users/sample-user/')

    def test_local_site_url_with_local_site(self):
        """Testing localsite's {% url %} with local site"""
        context = Context({
            'local_site_name': 'test',
        })

        t = Template('{% url "dashboard" %}')
        self.assertEqual(t.render(context), '/s/test/dashboard/')

        t = Template('{% url "user" "sample-user" %}')
        self.assertEqual(t.render(context), '/s/test/users/sample-user/')



"""Unit tests for reviewboard.accounts.search_indexes.

Version Added:
    8.0
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from django.contrib.auth.models import User
from django.db.models import Q
from django_assert_queries.testing import assert_queries

from reviewboard.accounts.models import Profile
from reviewboard.reviews.models import Group
from reviewboard.accounts.search_indexes import UserIndex
from reviewboard.search.testing import search_enabled
from reviewboard.site.models import LocalSite
from reviewboard.testing import TestCase

if TYPE_CHECKING:
    from collections.abc import Sequence

    from django_assert_queries import ExpectedQueries


class UserIndexTests(TestCase):
    """Unit tests for UserIndex.

    Version Added:
        8.0
    """

    def test_update(self) -> None:
        """Testing UserIndex.update"""
        users = self._create_users()

        group1 = self.create_review_group(name='group1')
        group1.users.add(*users[:5])

        group2 = self.create_review_group(name='group2')
        group2.users.add(*users[:-5])

        # Pre-populate caches.
        self.assertFalse(LocalSite.objects.has_local_sites())

        equeries: ExpectedQueries = [
            {
                'model': User,
                'only_fields': {
                    'email',
                    'first_name',
                    'last_name',
                    'username',
                    'profile__extra_data',
                    'profile__is_private',
                },
                'select_related': {
                    'profile',
                },
                'where': Q(is_active=True),
            },
            {
                'extra': {
                    '_prefetch_related_val_user_id': (
                        '"reviews_group_users"."user_id"',
                        [],
                    ),
                },
                'join_types': {
                    'reviews_group_users': 'INNER JOIN',
                },
                'model': Group,
                'num_joins': 1,
                'only_fields': {
                    'invite_only',
                    'name',
                },
                'tables': {
                    'reviews_group',
                    'reviews_group_users',
                },
                'where': Q(users__in=users),
            },
        ]

        with search_enabled():
            with assert_queries(equeries):
                UserIndex().update()

    def test_update_with_local_sites(self) -> None:
        """Testing UserIndex.update with Local Sites in the database
        """
        users = self._create_users()

        local_site = self.create_local_site(name='test-site')
        local_site.users.add(*users[:5],
                             *users[:-5])

        group1 = self.create_review_group(name='group1')
        group1.users.add(*users[5:10])

        group2 = self.create_review_group(name='group2',
                                          local_site=local_site)
        group2.users.add(*users[:-5])

        # Pre-populate caches.
        self.assertTrue(LocalSite.objects.has_local_sites())

        equeries: ExpectedQueries = [
            {
                'model': User,
                'only_fields': {
                    'email',
                    'first_name',
                    'last_name',
                    'username',
                    'profile__extra_data',
                    'profile__is_private',
                },
                'select_related': {'profile'},
                'where': Q(is_active=True),
            },
            {
                'extra': {
                    '_prefetch_related_val_user_id': (
                        '"reviews_group_users"."user_id"',
                        [],
                    ),
                },
                'join_types': {
                    'reviews_group_users': 'INNER JOIN',
                },
                'model': Group,
                'num_joins': 1,
                'only_fields': {
                    'invite_only',
                    'name',
                },
                'tables': {
                    'reviews_group',
                    'reviews_group_users',
                },
                'where': Q(users__in=users),
            },
            {
                'extra': {
                    '_prefetch_related_val_user_id': (
                        '"site_localsite_users"."user_id"',
                        [],
                    ),
                },
                'join_types': {
                    'site_localsite_users': 'INNER JOIN',
                },
                'model': LocalSite,
                'num_joins': 1,
                'only_fields': {
                    'name',
                },
                'tables': {
                    'site_localsite',
                    'site_localsite_users',
                },
                'where': Q(users__in=users),
            },
        ]

        with search_enabled():
            with assert_queries(equeries, with_tracebacks=True):
                UserIndex().update()

    def _create_users(self) -> Sequence[User]:
        """Create and return users to use for the tests.

        Returns:
            list of django.contrib.auth.models.User:
            The list of created users.
        """
        users: list[User] = []

        for i in range(20):
            user = self.create_user(username=f'user-{i:02}')
            users.append(user)

            if i > 10:
                # Create a profile.
                Profile.objects.create(
                    user=user,
                    is_private=(i % 2 == 0),
                )

        return users



from django.utils.html import escape
from djblets.markdown import markdown_escape, markdown_unescape
from djblets.webapi.resources.mixins.forms import (
    UpdateFormMixin as DjbletsUpdateFormMixin)

from reviewboard.reviews.markdown_utils import (markdown_set_field_escaped,
                                                render_markdown)
from reviewboard.webapi.base import ImportExtraDataError


class MarkdownFieldsMixin(object):
    """Mixes in common logic for Markdown text fields.

    Any resource implementing this is assumed to have at least one
    Markdown-capable text field.

    Clients can pass ``?force-text-type=`` (for GET) or ``force_text_type=``
    (for POST/PUT) with a value of ``plain``, ``markdown`` or ``html`` to
    return the given text fields in the payload using the requested format.

    When ``markdown`` is specified, the Markdown text fields will return valid
    Markdown content, escaping if necessary.

    When ``plain`` is specified, plain text will be returned instead. If
    the content was in Markdown before, this will unescape the content.

    When ``html`` is specified, the content will be transformed into HTML
    suitable for display.

    Clients can also pass ``?include-text-types=<type1>[,<type2>,...]``
    (for GET) or ``include_text_types=<type1>[,<type2>,...]`` (for POST/PUT)
    to return the text fields within special :samp:`{type}_text_fields`
    entries in the resource payload. A special type of "raw" is allowed,
    which will return the text types stored in the database.

    (Note that passing ``?include-text-types=raw`` is equivalent to passing
    ``?include-raw-text-fields=1`` in 2.0.9 and 2.0.10. The latter is
    deprecated.)
    """
    TEXT_TYPE_PLAIN = 'plain'
    TEXT_TYPE_MARKDOWN = 'markdown'
    TEXT_TYPE_HTML = 'html'
    TEXT_TYPE_RAW = 'raw'

    TEXT_TYPES = (TEXT_TYPE_PLAIN, TEXT_TYPE_MARKDOWN, TEXT_TYPE_HTML)
    SAVEABLE_TEXT_TYPES = (TEXT_TYPE_PLAIN, TEXT_TYPE_MARKDOWN)
    INCLUDEABLE_TEXT_TYPES = TEXT_TYPES + (TEXT_TYPE_RAW,)

    DEFAULT_EXTRA_DATA_TEXT_TYPE = TEXT_TYPE_MARKDOWN

    def serialize_object(self, obj, *args, **kwargs):
        """Serializes the object, transforming text fields.

        This is a specialization of serialize_object that transforms any
        text fields that support text types. It also handles attaching
        the raw text to the payload, on request.
        """
        data = super(MarkdownFieldsMixin, self).serialize_object(
            obj, *args, **kwargs)

        request = kwargs.get('request')

        if not request:
            force_text_type = None
        elif request.method == 'GET':
            force_text_type = request.GET.get('force-text-type')
        else:
            force_text_type = request.POST.get('force_text_type')

        if force_text_type not in self.TEXT_TYPES:
            force_text_type = None

        extra_text_type_fields = dict(
            (extra_text_type, {})
            for extra_text_type in self._get_extra_text_types(obj, **kwargs)
        )

        for field, field_info in self.fields.items():
            if not field_info.get('supports_text_types'):
                continue

            get_func = getattr(self, 'get_is_%s_rich_text' % field, None)

            if callable(get_func):
                getter = lambda obj, *args: get_func(obj)
            else:
                getter = lambda obj, data, rich_text_field, text_type_field: \
                    getattr(obj, rich_text_field, None)

            self._serialize_text_info(obj, data, extra_text_type_fields,
                                      field, force_text_type, getter)

        if 'extra_data' in data:
            extra_data = data['extra_data']
            all_text_types_extra_data = {}

            if obj.extra_data is None:
                obj.extra_data = {}

            # Work on a copy of extra_data, in case we change it.
            for field, value in obj.extra_data.copy().items():
                if not self.get_extra_data_field_supports_markdown(obj, field):
                    continue

                # If all_text_types_extra_data is empty that implies we have
                # encountered the first field in extra_data which supports
                # markdown. In this case we must initialize the dictionary
                # with the extra text types that should be included in the
                # payload.
                if not all_text_types_extra_data:
                    all_text_types_extra_data = {
                        _key: {}
                        for _key in extra_text_type_fields.keys()
                    }

                # Note that we assume all custom fields are in Markdown by
                # default. This is to preserve compatibility with older
                # fields. New fields will always have the text_type flag
                # set to the proper value.
                self._serialize_text_info(
                    obj, extra_data, all_text_types_extra_data, field,
                    force_text_type, self._extra_data_rich_text_getter)

            for key, values in all_text_types_extra_data.items():
                extra_text_type_fields[key]['extra_data'] = values

        for key, values in extra_text_type_fields.items():
            data[key + '_text_fields'] = values

        return data

    def _serialize_text_info(self, obj, data, extra_text_type_fields, field,
                             force_text_type, getter):
        text_type_field = self._get_text_type_field_name(field)
        rich_text_field = self._get_rich_text_field_name(field)

        field_is_rich_text = getter(obj, data, rich_text_field,
                                    text_type_field)
        assert field_is_rich_text is not None, \
            'No value for field "%s" found in %r' % (rich_text_field, obj)

        if field_is_rich_text:
            field_text_type = self.TEXT_TYPE_MARKDOWN
        else:
            field_text_type = self.TEXT_TYPE_PLAIN

        value = data.get(field)

        if force_text_type:
            data[text_type_field] = force_text_type

            if value is not None:
                data[field] = self._normalize_text(
                    value, field_is_rich_text, force_text_type)
        else:
            data[text_type_field] = field_text_type

        for extra_text_type in extra_text_type_fields:
            if extra_text_type == self.TEXT_TYPE_RAW:
                norm_extra_text_type = field_text_type
                norm_extra_value = value
            else:
                norm_extra_text_type = extra_text_type
                norm_extra_value = self._normalize_text(
                    value, field_is_rich_text, extra_text_type)

            extra_text_type_fields[extra_text_type].update({
                field: norm_extra_value,
                text_type_field: norm_extra_text_type,
            })

    def serialize_text_type_field(self, obj, request=None, **kwargs):
        return None

    def can_import_extra_data_field(self, obj, field):
        """Returns whether a particular field in extra_data can be imported.

        If an extra_data field is marked as supporting rich text, we'll skip
        importing it through normal means. Instead, it will be handled
        separately later.
        """
        return not self.get_extra_data_field_supports_markdown(obj, field)

    def get_extra_data_field_supports_markdown(self, obj, key):
        """Returns whether a particular field in extra_data supports Markdown.

        If the field supports Markdown text, the value will be normalized
        based on the requested ?force-text-type= parameter.
        """
        return False

    def _get_extra_text_types(self, obj, request=None, **kwargs):
        """Returns any extra text types that should be included in the payload.

        This will return the list of extra text types that can be included,
        filtering the list by those that are supported.

        It also checks for the older ``?include-raw-text-fields=1`` option,
        which is the same as using ``?include-text-types=raw``.
        """
        extra_text_types = set()

        if request:
            if request.method == 'GET':
                include_types = request.GET.get('include-text-types')
                include_raw_text = request.GET.get('include-raw-text-fields')
            else:
                include_types = request.POST.get('include_text_types')
                include_raw_text = request.POST.get('include_raw_text_fields')

            if include_raw_text in ('1', 'true'):
                extra_text_types.add(self.TEXT_TYPE_RAW)

            if include_types:
                extra_text_types.update([
                    text_type
                    for text_type in include_types.split(',')
                    if text_type in self.INCLUDEABLE_TEXT_TYPES
                ])

        return extra_text_types

    def _normalize_text(self, text, field_is_rich_text, force_text_type):
        """Normalizes text to the proper format.

        This considers the requested text format, and whether or not the
        value should be set for rich text.
        """
        assert force_text_type

        if text is not None:
            if force_text_type == self.TEXT_TYPE_PLAIN and field_is_rich_text:
                text = markdown_unescape(text)
            elif (force_text_type == self.TEXT_TYPE_MARKDOWN and
                  not field_is_rich_text):
                text = markdown_escape(text)
            elif force_text_type == self.TEXT_TYPE_HTML:
                if field_is_rich_text:
                    text = render_markdown(text)
                else:
                    text = escape(text)

        return text

    def set_text_fields(self, obj, text_field,
                        rich_text_field_name=None,
                        text_type_field_name=None,
                        text_model_field=None,
                        **kwargs):
        """Normalize Markdown-capable text fields that are being saved.

        Args:
            obj (django.db.models.Model):
                The object containing Markdown-capable fields to modify.

            text_field (unicode):
                The key from ``kwargs`` containing the new value for the
                text fields.

            rich_text_field_name (unicode, optional):
                The name of the boolean field on ``obj`` representing the
                rich text state.

                If not provided, a name will be generated based on the value
                of ``text_field`` (``'rich_text'`` if ``text_field`` is
                ``'text'``, or :samp:`{text_field}_rich_text` otherwise).

            text_type_field_name (unicode, optional):
                The key from ``kwargs`` containing the text type.

                If not provided, a name will be generated based on the value
                of ``text_field`` (``'text_type'`` if ``text_field`` is
                ``'text'``, or :samp:`{text_field}_text_type` otherwise).

            text_model_field (unicode, optional):
                The name of the text field on ``obj``.

                If not provided, ``text_field`` will be used for the value.

            **kwargs (dict):
                Any fields passed by the client. These should be values
                corresponding to ``text_type_field_name`` and ``text_field``
                in here.

                This may also contain a legacy value of ``'text_type'``,
                which will be used if ``text_type_field_name`` is not present.

        Returns:
            set:
            The fields on ``obj`` that were set based on the request.
        """
        modified_fields = set()

        if not text_model_field:
            text_model_field = text_field

        if not text_type_field_name:
            text_type_field_name = self._get_text_type_field_name(text_field)

        if not rich_text_field_name:
            rich_text_field_name = self._get_rich_text_field_name(text_field)

        old_rich_text = getattr(obj, rich_text_field_name, None)
        text = kwargs.get(text_field)

        if text is not None:
            setattr(obj, text_model_field, text.strip())
            modified_fields.add(text_model_field)

        legacy_text_type = kwargs.get('text_type')
        text_type = kwargs.get(text_type_field_name, legacy_text_type)

        if text_type is not None:
            rich_text = (text_type == self.TEXT_TYPE_MARKDOWN)

            setattr(obj, rich_text_field_name, rich_text)
            modified_fields.add(rich_text_field_name)

            # If the caller has changed the text type for this field, but
            # hasn't provided a new field value, then we will need to update
            # the affected field's existing contents by escaping or
            # unescaping.
            if rich_text != old_rich_text and text is None:
                markdown_set_field_escaped(obj, text_model_field, rich_text)
                modified_fields.add(text_model_field)
        elif old_rich_text:
            # The user didn't specify rich-text, but the object may be set
            # for rich-text, in which case we'll need to pre-escape the text
            # field.
            if text is not None:
                markdown_set_field_escaped(obj, text_model_field,
                                           old_rich_text)
                modified_fields.add(text_model_field)

        return modified_fields

    def set_extra_data_text_fields(self, obj, text_field, extra_fields,
                                   **kwargs):
        """Normalize Markdown-capable text fields in extra_data.

        This will check if any Markdown-capable text fields in extra_data
        have been changed (either by changing the text or the text type),
        and handle the saving of the text and type.

        This works just like :py:meth:`set_text_fields`, but specially handles
        how things are stored in extra_data (text_type vs. rich_text fields,
        possible lack of presence of a text_type field, etc.).

        Args:
            obj (django.db.models.Model):
                The object containing Markdown-capable fields to modify.

            text_field (unicode):
                The key from ``kwargs`` containing the new value for the
                text fields.

            extra_fields (dict):
                Fields passed to the API resource that aren't natively handled
                by that resource. These may contain ``extra_data.``-prefixed
                keys.

            **kwargs (dict):
                Any fields passed by the client. This may be checked for a
                legacy ``text_type`` field.

        Returns:
            set:
            The keys in ``extra_data`` that were set based on the request.
        """
        modified_fields = set()

        text_type_field = self._get_text_type_field_name(text_field)
        extra_data = obj.extra_data
        extra_data_text_field = 'extra_data.' + text_field
        extra_data_text_type_field = 'extra_data.' + text_type_field

        if extra_data_text_field in extra_fields:
            # This field was updated in this request. Make sure it's
            # stripped.
            extra_data[text_field] = \
                extra_fields[extra_data_text_field].strip()
            modified_fields.add(text_field)
        elif extra_data_text_type_field not in extra_fields:
            # Nothing about this field has changed, so bail.
            return modified_fields

        old_text_type = extra_data.get(text_type_field)
        text_type = extra_fields.get(extra_data_text_type_field,
                                     kwargs.get('text_type'))

        if text_type is not None:
            if old_text_type is None:
                old_text_type = self.DEFAULT_EXTRA_DATA_TEXT_TYPE

            # If the caller has changed the text type for this field, but
            # hasn't provided a new field value, then we will need to update
            # the affected field's existing contents by escaping or
            # unescaping.
            if (text_type != old_text_type and
                extra_data_text_field not in extra_fields):
                markdown_set_field_escaped(
                    extra_data, text_field,
                    text_type == self.TEXT_TYPE_MARKDOWN)
                modified_fields.add(text_field)
        elif old_text_type:
            # The user didn't specify a text type, but the object may be set
            # for Markdown, in which case we'll need to pre-escape the text
            # field.
            if extra_data_text_field in extra_fields:
                markdown_set_field_escaped(
                    extra_data, text_field,
                    old_text_type == self.TEXT_TYPE_MARKDOWN)
                modified_fields.add(text_field)

        # Ensure we have a text type set for this field. If one wasn't
        # provided or set above, we'll set it to the default now.
        extra_data[text_type_field] = \
            text_type or self.DEFAULT_EXTRA_DATA_TEXT_TYPE
        modified_fields.add(text_type_field)

        return modified_fields

    def _get_text_type_field_name(self, text_field_name):
        if text_field_name == 'text':
            return 'text_type'
        else:
            return '%s_text_type' % text_field_name

    def _get_rich_text_field_name(self, text_field_name):
        if text_field_name == 'text':
            return 'rich_text'
        else:
            return '%s_rich_text' % text_field_name

    def _extra_data_rich_text_getter(self, obj, data, rich_text_field,
                                     text_type_field):
        text_type = data.get(text_type_field,
                             self.DEFAULT_EXTRA_DATA_TEXT_TYPE)

        return text_type == self.TEXT_TYPE_MARKDOWN


class UpdateFormMixin(DjbletsUpdateFormMixin):
    """A mixin for providing the ability to create and update using a form.

    A WebAPIResource class using this mixin must set the :py:attr:`form_class`
    attribute to a :py:class:`ModelForm` instance that corresponds to the model
    being updated.

    Classes using this mixin can provide methods of the form
    ``parse_<field_name>_field`` to do parsing of form data before it is passed
    to the form. These methods may return either a single value or a list (in
    the case where the corresponding field expects a list of values, such as a
    :py:class:`django.forms.ModelMultipleChoiceField`).

    The :py:meth:`create_form` and :py:meth:`save_form` methods should be used
    for creating new form instances and saving them. A form created this way
    can be given an optional instance argument to allow for updating the
    instance. Any fields missing from data (but appearing in the
    :py:class:`form_class`'s :py:attr:`fields` attribute) will be copied over
    from the instance.
    """

    def handle_form_request(self, **kwargs):
        """Handle an HTTP request for creating or updating through a form.

        This simply wraps the parent method and handles
        :py:class:`~reviewboard.webapi.base.ImportExtraDataError` exceptions
        during form save.

        Args:
            **kwargs (dict):
                Keyword arguments to pass to the parent method.

        Returns:
            tuple or django.http.HttpResponse:
            The response to send back to the client.
        """
        try:
            return super(UpdateFormMixin, self).handle_form_request(**kwargs)
        except ImportExtraDataError as e:
            return e.error_payload

    def save_form(self, form, save_kwargs, extra_fields=None, **kwargs):
        """Save the form and extra data.

        Args:
            form (django.forms.ModelForm):
                The form to save.

            save_kwargs (dict):
                Additional keyword arguments to pass to the
                :py:class:`ModelForm.save() <django.forms.ModelForm.save>`.

            extra_fields (dict, optional):
                The extra data to save on the object. These should be key-value
                pairs in the form of ``extra_data.key = value``.

            **kwargs (dict):
                Additional keyword arguments to pass to the parent method.

        Returns:
            django.db.models.Model:
            The saved model instance.

        Raises:
            reviewboard.webapi.base.ImportExtraDataError:
                Extra data failed to import. The form will not be saved.
        """
        if save_kwargs:
            save_kwargs = save_kwargs.copy()
        else:
            save_kwargs = {}

        save_kwargs['commit'] = False

        instance = super(UpdateFormMixin, self).save_form(
            form=form,
            save_kwargs=save_kwargs,
            **kwargs)

        if extra_fields:
            if not instance.extra_data:
                instance.extra_data = {}

            self.import_extra_data(instance, instance.extra_data, extra_fields)

        instance.save()
        form.save_m2m()

        return instance



"""Information on product licenses.

Version Added:
    8.0
"""

from __future__ import annotations

import logging
from enum import Enum
from typing import TYPE_CHECKING

from django.utils import timezone
from django.utils.translation import (gettext_lazy as _,
                                      ngettext_lazy as N_)
from djblets.util.typing import StrOrPromise
from typing_extensions import TypedDict

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from datetime import datetime
    from typing import Any

    from typing_extensions import NotRequired


logger = logging.getLogger(__name__)


class LicenseStatus(Enum):
    """The status of a license.

    Version Added:
        8.0
    """

    #: The license has not yet been activated.
    UNLICENSED = 'unlicensed'

    #: The license is applied and active.
    LICENSED = 'licensed'

    #: The license has expired and is in a grace period.
    EXPIRED_GRACE_PERIOD = 'expired-grace-period'

    #: The license has expired and is past its grace period.
    HARD_EXPIRED = 'hard-expired'


class LicenseLineItem(TypedDict):
    """A line item to show for a license.

    Version Added:
        8.0
    """

    #: The content for the line item.
    #:
    #: This may be a plain text string (which will be escaped) or a safe
    #: HTML-formatted string, depending on the :py:attr:`content_is_html`
    #: flag.
    content: str

    #: Whether the content should be rendered as HTML.
    content_is_html: NotRequired[bool]

    #: The optional icon CSS class name to display alongside the content.
    icon: NotRequired[str]


class LicenseInfo:
    """Information on a license.

    This may be an active, expired, or trial license.

    License providers should use this to communicate information about a
    license to the rest of Review Board, but generally will have their own
    more tailored license structure behind it. This is not intended as a
    fully-featured license model.

    Version Added:
        8.0
    """

    ######################
    # Instance variables #
    ######################

    #: Whether the license is set to auto-renew on the license server.
    #:
    #: When ``True``, the license is presented as renewing rather than
    #: expiring. The license summary and detail line will say "renews on"
    #: instead of "expires on", and we don't show warnings that the license
    #: will expire soon.
    #:
    #: This is dynamic billing-side state and is not part of the signed
    #: license payload. License providers should populate this from the
    #: most recent license check.
    auto_renew: bool

    #: Whether a new license file can be manually uploaded for this license.
    can_upload_license: bool

    #: The expiration date/time of the license.
    expires: datetime | None

    #: The number of grace period days remaining on the license.
    #:
    #: This is only considered when the license is expired.
    grace_period_days_remaining: int

    #: Whether this is a trial license.
    is_trial: bool

    #: A backend-specific ID for this license.
    #:
    #: This may be used for communication with a license server,
    #: client-side.
    license_id: str

    #: The backend-specific license instance.
    #:
    #: This is used purely for the convenience of a license provider.
    #: It may be ``None``.
    license_instance: Any

    #: Who the license is licensed to.
    #:
    #: This may be the company or a division of a company (cost center).
    licensed_to: str

    #: Any displayable line items to show on license information.
    line_items: Sequence[LicenseLineItem]

    #: A backend-specific plan ID for this license.
    plan_id: str | None

    #: A display name for the plan.
    plan_name: str | None

    #: The name of the product being licensed.
    product_name: str

    #: The active/expiration status of this license.
    status: LicenseStatus

    #: A descriptive summary of the license.
    #:
    #: If not provided, one will be automatically generated based off the
    #: product name and status.
    summary: str

    def __init__(
        self,
        *,
        license_id: str,
        licensed_to: str,
        product_name: str,
        auto_renew: bool = False,
        can_upload_license: bool = False,
        expires: (datetime | None) = None,
        grace_period_days_remaining: int = 0,
        is_trial: bool = False,
        license_instance: Any = None,
        line_items: (Sequence[LicenseLineItem] | None) = None,
        plan_id: (str | None) = None,
        plan_name: (str | None) = None,
        status: LicenseStatus = LicenseStatus.UNLICENSED,
        summary: str = '',
    ) -> None:
        """Initialize the license information.

        Args:
            licensed_id (str):
                A backend-specific ID for this license.

                This may be used for communication with a license server,
                client-side.

            licensed_to (str):
                Who the license is licensed to.

                This may be the company or a division of a company (cost
                center).

            product_name (str):
                The name of the product being licensed.

            auto_renew (bool, optional):
                Whether the license is set to auto-renew on the license server.
                When ``True``, the license summary and details show "renews on"
                instead of "expires on".

            can_upload_license (bool, optional):
                Whether a new license file can be manually uploaded for
                this license.

            expires (datetime.datetime, optional):
                The expiration date/time of the license.

            grace_period_days_remaining (int, optional):
                The number of grace period days remaining on the license.

                This is only considered when the license is expired.

            is_trial (bool, optional):
                Whether this is a trial license.

            license_instance (object, optional):
                The backend-specific license instance.

                This is used purely for the convenience of a license provider.
                It may be ``None``.

            line_items (Sequence[LicenseLineItem], optional):
                Any displayable line items to show on license information.

            plan_id (str, optional):
                A backend-specific plan ID for this license.

            plan_name (str, optional):
                A display name for the plan.

            status (LicenseStatus, optional):
                The active/expiration status of this license.

            summary (str, optional):
                A descriptive summary of the license.

                If not provided, one will be automatically generated based
                off the product name and status.
        """
        self.license_id = license_id
        self.licensed_to = licensed_to
        self.product_name = product_name
        self.auto_renew = auto_renew
        self.can_upload_license = can_upload_license
        self.expires = expires
        self.grace_period_days_remaining = grace_period_days_remaining
        self.is_trial = is_trial
        self.license_instance = license_instance
        self.line_items = line_items or []
        self.plan_id = plan_id
        self.plan_name = plan_name
        self.status = status
        self.summary = summary

    def get_summary(self) -> str:
        """Return a summary for the license.

        This will return :py:attr:`summary` if provided. Otherwise, it
        will generate a suitable summary based on the state of the license.

        Subclasses can override this to provide custom summaries.

        Returns:
            str:
            The summary of the license.
        """
        summary: StrOrPromise = self.summary

        if summary:
            assert isinstance(summary, str)

            return summary

        status = self.status

        key: list[str] = []
        fmt_args: dict[str, object] = {
            'product': self.product_name,
        }

        if status != LicenseStatus.UNLICENSED:
            is_trial = self.is_trial
            plan_name = self.plan_name
            expires = self.expires

            if is_trial:
                key.append('trial')

            if plan_name:
                key.append('plan')
                fmt_args['plan'] = self.format_plan_name()

            if status == LicenseStatus.LICENSED:
                if expires is not None:
                    fmt_args['days_remaining'] = \
                        (expires - timezone.now()).days

                    if not is_trial and self.get_expires_soon():
                        # When the license is set to auto-renew, frame
                        # the upcoming date as a renewal instead of an
                        # expiration.
                        if self.auto_renew:
                            key.append('auto_renew')
                        else:
                            key.append('expires_soon')
                else:
                    assert not is_trial
            elif status == LicenseStatus.EXPIRED_GRACE_PERIOD:
                fmt_args['days_remaining'] = self.grace_period_days_remaining

        try:
            summary = _DEFAULT_SUMMARY_FORMATS[status][tuple(key)]
        except KeyError:
            logger.error('Hit a bad license summary state for license %r: '
                         'Missing string for status=%r, key %r',
                         self, status, tuple(key))

            summary = _("{product}'s license status is unknown")

        return summary.format(**fmt_args)

    def format_plan_name(self) -> str:
        """Return a formatted version of the plan name.

        This allows a license implementation to provide a custom way of
        formatting the plan name for use in the summary.

        By default, this returns the plan name as-is.

        This may not be used if a plan name is not set.

        Returns:
            str:
            The formatted plan name.
        """
        plan_name = self.plan_name
        assert plan_name

        return plan_name

    def get_expires_soon(self) -> bool:
        """Return whether or not the license expires soon.

        "Soon" is defined as the result of
        :py:meth:`get_expires_soon_days_threshold`, which can be overridden
        by subclasses. By default, this is defined as within 10 days for a
        trial license or within 30 days for a non-trial license.

        This does not include the grace period.

        Returns:
            bool:
            ``True`` if the license expires soon. ``False`` if it does not,
            or if the license has already expired.
        """
        if self.status != LicenseStatus.LICENSED or not self.expires:
            return False

        days_threshold = self.get_expires_soon_days_threshold()
        days_remaining = (self.expires - timezone.now()).days

        return days_remaining <= days_threshold

    def get_expires_soon_days_threshold(self) -> int:
        """Return the number of days considered "soon" for expiration.

        This can be overridden by subclasses to change the number of days
        based on the term or renewal times of a license.

        By default, this is defined as within 10 days for a trial license or
        within 30 days for a non-trial license.

        This does not include the grace period.

        Returns:
            int:
            The number of days considered "soon" for expiration, based on
            the license state or type.
        """
        if self.is_trial:
            return 10
        else:
            return 30

    def __repr__(self) -> str:
        """Return a representation of this license.

        Returns:
            str:
            The representation.
        """
        cls_name = type(self).__name__

        return (
            f'<{cls_name}'
            f'(license_id={self.license_id!r},'
            f' status={self.status!r},'
            f' product_name={self.product_name!r},'
            f' plan_id={self.plan_id!r},'
            f' plan_name={self.plan_name!r},'
            f' licensed_to={self.licensed_to!r},'
            f' is_trial={self.is_trial},'
            f' expires={self.expires!r},'
            f' grace_period_days_remaining={self.grace_period_days_remaining},'
            f' line_items={self.line_items!r}>'
        )


if TYPE_CHECKING:
    _DefaultSummaryFormats = Mapping[
        LicenseStatus,
        Mapping[tuple[str, ...], StrOrPromise],
    ]


#: Strings for the default license summaries, based on state.
#:
#: This provides a single place to look up the default summaries for a
#: license based on the trial/purchased state, presence of a plan, and
#: whether there are optional days remaining to display for the license
#: term.
#:
#: Version Added:
#:     8.0
_DEFAULT_SUMMARY_FORMATS: _DefaultSummaryFormats = {
    LicenseStatus.UNLICENSED: {
        (): _('{product} is not licensed!'),
    },

    LicenseStatus.LICENSED: {
        # Purchased license is active
        (): _('License for {product} is active'),

        # Purchased license with plan is active
        ('plan',): _('License for {product} {plan} is active'),

        # Purchased license expires soon
        ('expires_soon',): N_(
            ('License for {product} expires in {days_remaining} day'),
            ('License for {product} expires in {days_remaining} days'),
            'days_remaining'
        ),

        # Purchased license with plan expires soon
        ('plan', 'expires_soon'): N_(
            ('License for {product} {plan} expires in {days_remaining} '
             'day'),
            ('License for {product} {plan} expires in {days_remaining} '
             'days'),
            'days_remaining'
        ),

        # Purchased license auto-renews soon
        ('auto_renew',): N_(
            ('License for {product} renews in {days_remaining} day'),
            ('License for {product} renews in {days_remaining} days'),
            'days_remaining'
        ),

        # Purchased license with plan auto-renews soon
        ('plan', 'auto_renew'): N_(
            ('License for {product} {plan} renews in {days_remaining} '
             'day'),
            ('License for {product} {plan} renews in {days_remaining} '
             'days'),
            'days_remaining'
        ),

        # Trial license is active
        ('trial',): N_(
            'Trial license for {product} ends in {days_remaining} day',
            'Trial license for {product} ends in {days_remaining} days',
            'days_remaining',
        ),

        # Trial license with plan is active
        ('trial', 'plan'): N_(
            'Trial license for {product} {plan} ends in {days_remaining} day',
            'Trial license for {product} {plan} ends in {days_remaining} days',
            'days_remaining',
        ),
    },

    LicenseStatus.EXPIRED_GRACE_PERIOD: {
        # Purchased license expired with grace period
        (): N_(
            ('License for {product} expired and will stop working in '
             '{days_remaining} day'),
            ('License for {product} expired and will stop working in '
             '{days_remaining} days'),
            'days_remaining'
        ),

        # Purchased license with plan expired with grace period
        ('plan',): N_(
            ('License for {product} {plan} expired and will stop working '
             'in {days_remaining} day'),
            ('License for {product} {plan} expired and will stop working '
             'in {days_remaining} days'),
            'days_remaining'
        ),

        # Trial license expired with grace period
        ('trial',): N_(
            ('Trial license for {product} expired and will stop working in '
             '{days_remaining} day'),
            ('Trial license for {product} expired and will stop working in '
             '{days_remaining} days'),
            'days_remaining'
        ),

        # Trial license with plan expired with grace period
        ('trial', 'plan'): N_(
            ('Trial license for {product} {plan} expired and will stop '
             'working in {days_remaining} day'),
            ('Trial license for {product} {plan} expired and will stop '
             'working in {days_remaining} days'),
            'days_remaining'
        ),
    },

    LicenseStatus.HARD_EXPIRED: {
        # Purchased license hard-expired
        (): _('License for {product} expired and needs to be renewed'),

        # Purchased license with plan hard-expired
        ('plan',): _(
            'License for {product} {plan} expired and needs to be renewed'
        ),

        # Trial license hard-expired
        ('trial',): _('Trial license for {product} expired'),

        # Trial license with plan hard-expired
        ('trial', 'plan'): _('Trial license for {product} {plan} expired'),
    },
}



"""Command line tool for helping develop extensions.

This tool, :command:`rbext`, currently provides the ability to easily run
extension-provided unit tests. In the future, it'll also help with other
development tasks, and with updating the Package Store with the latest versions
of an extension.
"""

from __future__ import annotations

import argparse
import logging
import os
import re
import sys
from importlib import import_module
from textwrap import dedent
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import NoReturn


logger = logging.getLogger(__name__)


os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'reviewboard.settings')

rbext_dir = os.path.abspath(os.path.dirname(__file__))
sys.path.insert(0, os.path.join(rbext_dir, 'conf', 'rbext'))

from django.utils.encoding import force_str
from importlib_metadata import Prepared

from reviewboard import (
    __version__ as reviewboard_version,
    get_manual_url,
)
from reviewboard.cmdline.utils.argparsing import (HelpFormatter,
                                                  RBProgVersionAction)
from reviewboard.cmdline.utils.console import init_console


# NOTE: We want to include Django-based modules as late as possible, in order
#       to allow extension-provided settings to apply.


MANUAL_URL = get_manual_url()
EXTENSION_MANUAL_URL = f'{MANUAL_URL}extending/'

console = None


class BaseCommand:
    """Base class for a command."""

    #: The name of a command.
    #:
    #: This is what the user will type on the command line after ``rbext``.
    name = None

    #: The summary of the command.
    #:
    #: This will be shown in the help output.
    help_summary = 'Run unit tests for an extension.'

    #: A description of the command, when displaying the command's own help.
    description_text = None

    def add_options(
        self,
        parser: argparse.ArgumentParser,
    ) -> None:
        """Add any command-specific options to the parser.

        Args:
            parser (argparse.ArgumentParser):
                The argument parser.
        """
        pass

    def run(
        self,
        options: argparse.Namespace,
    ) -> int:
        """Run the command.

        This will perform common setup work and then hand things off to
        :py:meth:`main`.

        Args:
            options (argparse.Namespace):
                Options set from the arguments.

        Returns:
            int:
            The return code for the process.
        """
        if options.settings_file:
            sys.path.insert(
                0,
                os.path.abspath(os.path.dirname(options.settings_file)))

        return self.main(options)

    def main(
        self,
        options: argparse.Namespace,
    ) -> int:
        """Perform the main operations for the command.

        Args:
            options (argparse.Namespace):
                Options set from the arguments.

        Returns:
            int:
            The command's exit code.
        """
        raise NotImplementedError

    def error(
        self,
        msg: str,
    ) -> NoReturn:
        """Display a fatal error to the user and exit.

        Args:
            msg (str):
                The message to display.

        Raises:
            django.core.management.CommandError:
            The resulting error.
        """
        from django.core.management import CommandError

        raise CommandError(msg)


class TestCommand(BaseCommand):
    """A command that runs an extension's test suite."""

    name = 'test'
    help_summary = 'Run unit tests for an extension.'
    description_text = (
        'Runs the unit tests for your extension, looking for any tests.py '
        'or tests/*.py files.\n'
        '\n'
        'For simple extensions, you can just provide -m <modulename>.\n'
        '\n'
        'For more advanced extensions with many apps, you may need to '
        'define a custom settings_local.py file in your tree based on your '
        'Review Board development environment, RB_EXTRA_APPS to a list of '
        'Django app names required by the extension.\n'
        '\n'
        'This will wrap the "nose" test runner, which provides many '
        'additional options for running your tests. To see all available '
        'options, run:\n'
        '\n'
        '  nosetests --help\n'
        '\n'
        'To use any additional "nose" options, run this command with a '
        '"--" before the arguments and test names:\n'
        '\n'
        '  rbext test <options> -- <nose-options> <tests>'
    )

    def add_options(
        self,
        parser: argparse.ArgumentParser,
    ) -> None:
        """Add command line arguments for running tests.

        Args:
            parser (argparse.ArgumentParser):
                The argument parser.
        """
        parser.add_argument(
            '--pytest',
            action='store_true',
            default=False,
            dest='pytest',
            help='Use pytest runner instead of nose.')
        parser.add_argument(
            '--app',
            metavar='APP_LABEL',
            action='append',
            dest='app_names',
            help='A Django app label to add to the list of installed apps. '
                 'This is only required for tests that use apps not '
                 'enabled by extensions.')
        parser.add_argument(
            '-e',
            '--extension',
            metavar='EXTENSION_CLASS',
            dest='extension_class',
            help='The full module and class path to the extension to test.')
        parser.add_argument(
            '-m',
            '--module',
            action='append',
            metavar='MODULE_NAME',
            dest='module_names',
            help='The name(s) of the extension module(s) to test. For '
                 'example, if your tests are in "myextension.tests", you '
                 'might want to use "myextension". This may require '
                 'specifying multiple modules in the extension, and any '
                 'dependencies. You may want to use --extension instead.')
        parser.add_argument(
            '--pdb',
            action='append_const',
            dest='test_options',
            const='--pdb',
            help='Drop into a debugger on any failures or errors.')
        parser.add_argument(
            '--tree-root',
            metavar='PATH',
            default=os.getcwd(),
            help='The path to the root of the source tree.')
        parser.add_argument(
            '--with-coverage',
            action='append_const',
            dest='test_options',
            const='--with-coverage',
            help='Display a report on code covered or missed by tests.')
        parser.add_argument(
            '-x',
            '--stop',
            action='append_const',
            dest='test_options',
            const='-x',
            help='Stop running tests after the first failure.')
        parser.add_argument(
            'tests',
            metavar='TEST',
            nargs='*',
            help='Specific tests to run. This can be in the form of '
                 'mypackage.mymodule, mypackage.mymodule:TestsClass, or '
                 'mypackage.mymodule:TestsClass.test_method.')

    def main(
        self,
        options: argparse.Namespace,
    ) -> int:
        """Main function for running unit tests for the extension.

        Args:
            options (argparse.Namespace):
                Options set from the arguments.

        Returns:
            int:
            The command's exit code.
        """
        module_names = options.module_names or []

        os.environ['RB_TEST_MODULES'] = force_str(','.join(module_names))

        os.chdir(options.tree_root)
        os.environ['RB_RUNNING_TESTS'] = '1'

        from django import setup
        from django.apps import apps
        from django.conf import settings

        if not apps.ready:
            setup()

        installed_apps = list(settings.INSTALLED_APPS)

        # If an explicit extension is specified, then we'll want to grab its
        # list of apps.
        extension_class_name = options.extension_class

        assert console is not None

        if extension_class_name:
            module_name, class_name = extension_class_name.rsplit('.', 1)

            try:
                extension_class = getattr(import_module(module_name),
                                          class_name)
            except AttributeError:
                console.error(
                    f'The provided extension class "{class_name}" could not '
                    f'be found in {module_name}'
                )

                return 1
            except ImportError:
                console.error(
                    f'The provided extension class module "{module_name}" '
                    f'could not be found'
                )

                return 1

            installed_apps += (extension_class.apps or
                               [module_name.rsplit('.', 1)[0]])

        if options.app_names:
            installed_apps += options.app_names

        if installed_apps != list(settings.INSTALLED_APPS):
            settings.INSTALLED_APPS = installed_apps
            apps.set_installed_apps(installed_apps)

        from reviewboard.test import RBTestRunner

        use_pytest = options.pytest or os.path.exists('conftest.py')

        if not use_pytest:
            console.note(
                'Tests are running using the legacy nose test runner. Review '
                'Board 7 will switch to a pytest-based runner. To opt in to '
                'the new behavior, run with --pytest or create a conftest.py '
                'file.')

        test_runner = RBTestRunner(
            test_packages=module_names,
            cover_packages=module_names,
            verbosity=1,
            needs_collect_static=False,
            use_pytest=use_pytest)

        # Don't use +=, as we don't want to modify the list on the class.
        # We want to create a new one on the instance.
        test_runner.nose_options += (options.test_options or [])

        failures = test_runner.run_tests(options.tests)

        if failures:
            return 1
        else:
            return 0


class CreateCommand(BaseCommand):
    """A command for creating a new extension package."""

    name = 'create'
    help_summary = 'Create a new extension source tree.'
    description_text = (
        f'This takes care of creating a boilerplate extension and source '
        f'tree, giving you a starting point for developing a new extension.\n'
        f'\n'
        f'See the documentation for creating extensions:\n'
        f'\n'
        f'{EXTENSION_MANUAL_URL}'
    )

    def add_options(
        self,
        parser: argparse.ArgumentParser,
    ) -> None:
        """Add command line arguments for creating an extension.

        Args:
            parser (argparse.ArgumentParser):
                The argument parser.
        """
        parser.add_argument(
            '--name',
            required=True,
            help='The human-readable name for the extension. This is '
                 'required.')
        parser.add_argument(
            '--class-name',
            default=None,
            help='The class name for the extension (generally in CamelCase '
                 'form, without spaces). If not provided, this will be '
                 'based on the extension name.')
        parser.add_argument(
            '--package-name',
            default=None,
            help='The name of the package (using alphanumeric  ). '
                 'If not provided, this will be based on the extension '
                 'name.')
        parser.add_argument(
            '--package-version',
            default='1.0',
            help='The version for your extension and package.')
        parser.add_argument(
            '--summary',
            default=None,
            help='A one-line summary of the extension.')
        parser.add_argument(
            '--description',
            default=None,
            help='A short description of the extension.')
        parser.add_argument(
            '--author-name',
            default=None,
            help='The name of the author for the package and extension '
                 'metadata. This can be a company name.')
        parser.add_argument(
            '--author-email',
            default=None,
            help='The e-mail address of the author for the package and '
                 'extension metadata.')
        parser.add_argument(
            '--enable-configuration',
            action='store_true',
            default=False,
            help='Whether to enable a Configure button and view for the '
                 'extension.')
        parser.add_argument(
            '--enable-static-media',
            action='store_true',
            default=False,
            help='Whether to enable static media files for the package.')

    def main(
        self,
        options: argparse.Namespace,
    ) -> int:
        """Main function for creating an extension.

        Args:
            options (argparse.Namespace):
                Options set from the arguments.

        Returns:
            int:
            The command's exit code.
        """
        self._process_options(options)

        name = options.name
        package_name = options.package_name
        summary = options.summary
        description = options.description
        class_name = options.class_name
        configurable = options.enable_configuration
        enable_static_media = options.enable_static_media

        # Create the directory hierarchy.
        root_dir = package_name

        if os.path.exists(root_dir):
            self.error(
                f'There\'s already a directory named "{root_dir}". You must '
                f'remove it before you can create a new extension there.'
            )

        ext_dir = os.path.join(root_dir, package_name)
        static_dir = os.path.join(ext_dir, 'static')
        templates_dir = os.path.join(ext_dir, 'templates')

        for path in (root_dir, ext_dir):
            os.mkdir(path, 0o755)

        if enable_static_media:
            os.mkdir(static_dir, 0o755)

            for path in ('css', 'js', 'images'):
                os.mkdir(os.path.join(static_dir, path))

        # Create the packaging files.
        self._write_file(
            os.path.join(root_dir, 'README.md'),
            self._create_readme(name=name,
                                summary=summary,
                                description=description))

        self._write_file(
            os.path.join(root_dir, 'MANIFEST.in'),
            self._create_manifest(static_dir=static_dir,
                                  templates_dir=templates_dir))

        self._write_file(
            os.path.join(root_dir, 'pyproject.toml'),
            self._create_pyproject_toml(author=options.author_name,
                                        author_email=options.author_email,
                                        class_name=class_name,
                                        package_name=package_name,
                                        summary=summary,
                                        version=options.package_version))

        self._write_file(
            os.path.join(root_dir, 'conftest.py'),
            self._create_conftest_py())

        # Create the extension source files.
        self._write_file(os.path.join(ext_dir, '__init__.py'), '')

        self._write_file(
            os.path.join(ext_dir, 'extension.py'),
            self._create_extension_py(
                name=name,
                package_name=package_name,
                class_name=class_name,
                summary=summary,
                configurable=configurable,
                has_static_media=enable_static_media))

        if configurable:
            form_class_name = f'{class_name}Form'

            self._write_file(
                os.path.join(ext_dir, 'admin_urls.py'),
                self._create_admin_urls_py(
                    package_name=package_name,
                    class_name=class_name,
                    form_class_name=form_class_name))

            self._write_file(
                os.path.join(ext_dir, 'forms.py'),
                self._create_forms_py(form_class_name=form_class_name))

        # We're done!
        assert console is not None
        console.print(
            f'Generated a new extension in {os.path.abspath(root_dir)}')
        console.print()
        console.print('For information on writing your extension, see')
        console.print(EXTENSION_MANUAL_URL)

        return 0

    def _process_options(
        self,
        options: argparse.Namespace,
    ) -> None:
        """Process and normalize any provided options.

        This will attempt to provide suitable defaults for missing parameters,
        and to check that others are valid.

        Args:
            options (argparse.Namespace):
                Options set from the arguments.
        """
        name = options.name.strip()
        package_name = options.package_name
        class_name = options.class_name

        assert console is not None

        if not package_name:
            package_name = self._normalize_package_name(name)
            console.print(f'Using "{package_name}" as the package name.')
        else:
            package_name = package_name.strip()

            if not re.match(r'[A-Za-z][A-Za-z0-9._-]*', package_name):
                normalized = self._normalize_package_name(package_name)
                self.error(
                    f'"{package_name}" is not a valid package name. Try '
                    f'--package-name="{normalized}"'
                )

        if not class_name:
            class_name = self._normalize_class_name(name)
            console.print(f'Using "{class_name}" as the extension class name.')
        else:
            class_name = class_name.strip()

            if not re.match(r'[A-Za-z][A-Za-z0-9_]+Extension$', class_name):
                normalized = self._normalize_class_name(class_name)
                self.error(
                    f'"{class_name}" is not a valid class name. Try '
                    f'--class-name="{normalized}"'
                )

        options.name = name
        options.package_name = package_name
        options.class_name = class_name

    def _normalize_package_name(
        self,
        name: str,
    ) -> str:
        """Normalize a package name.

        This will ensure the package name is in a suitable format, replacing
        any invalid characters or dashes with ``_``, and converting it to
        lowercase.

        Args:
            name (str):
                The name of the package to normalize.

        Returns:
            str:
            The normalized name.
        """
        return Prepared.normalize(name)

    def _normalize_class_name(
        self,
        name: str,
    ) -> str:
        """Normalize a class name.

        This will ensure the class name is in a suitable format, converting it
        to CamelCase and adding an "Extension" at the end if needed.

        Args:
            name (str):
                The name of the class to normalize.

        Returns:
            str:
            The normalized class name.
        """
        class_name = ''.join(
            word.capitalize()
            for word in re.sub(r'[^A-Za-z0-9]+', ' ', name).split()
        )

        if not class_name.endswith('Extension'):
            class_name += 'Extension'

        return class_name

    def _write_file(
        self,
        filename: str,
        content: str,
        mode: (int | None) = None,
    ) -> None:
        """Write content to a file.

        This will create the file and write the provided content. The content
        will be stripped and dedented, with a trailing newline added, allowing
        the generation code to make use of multi-line strings.

        Args:
            filename (str):
                The path of the file to write.

            content (str):
                The content to write.

            mode (int):
                The optional permissions mode to set for the file.
        """
        with open(filename, 'w', encoding='utf-8') as fp:
            fp.write(dedent(content).strip())
            fp.write('\n')

        if mode is not None:
            os.chmod(filename, mode)

    def _create_readme(
        self,
        name: str,
        summary: str,
        description: str,
    ) -> str:
        """Create the content for a README.rst file.

        Args:
            name (str):
                The extension's name.

            summary (str):
                The extension's summary.

            description (str):
                A description of the extension.

        Returns:
            str:
            The resulting content for the file.
        """
        header_bar = '=' * len(name)
        content = '\n\n'.join(
            content
            for content in (summary, description)
            if content
        ) or 'Describe your extension.'

        return f"""
            {name}
            {header_bar}

            {content}
        """

    def _create_manifest(
        self,
        templates_dir: str,
        static_dir: str,
    ) -> str:
        """Create the content for a MANIFEST.in file.

        Args:
            templates_dir (str):
                The relative path to the templates directory.

            static_dir (str):
                The relative path to the static media directory.

        Returns:
            str:
            The resulting content for the file.
        """
        return f"""
            graft {templates_dir}
            graft {static_dir}

            include COPYING
            include INSTALL
            include README.md
            include *-requirements.txt

            global-exclude .*.sw[op] *.py[co] __pycache__ .DS_Store .noseids
        """

    def _create_pyproject_toml(
        self,
        *,
        author: str,
        author_email: str,
        class_name: str,
        package_name: str,
        summary: str,
        version: str,
    ) -> str:
        """Create the content for a pyproject.toml file.

        Version Changed:
            8.0:
            Added ``author``, ``author_email``, ``class_name``, ``summary``,
            and ``version`` arguments.

        Args:
            author (str):
                The name of the author of the extension.

                Version Added:
                    8.0

            author_email (str):
                The e-mail address of the author of the extension.

                Version Added:
                    8.0

            class_name (str):
                The name of the extension class.

                Version Added:
                    8.0

            package_name (str):
                The name of the package.

            summary (str):
                A summary of the package.

                Version Added:
                    8.0

            version (str):
                The version of the package.

                Version Added:
                    8.0

        Returns:
            str:
            The resulting content for the file.
        """
        if not author:
            author = '<REPLACE ME>'

        if not author_email:
            author_email = '<REPLACE ME>'

        if not summary:
            summary = '<REPLACE ME>'

        return f"""
            [build-system]
            requires = [
                # Update this for the target version of Review Board.
                'reviewboard~={reviewboard_version}',

                'reviewboard[extension-packaging]',
            ]
            build-backend = 'reviewboard.extensions.packaging.backend'


            [project]
            name = '{package_name}'
            version = '{version}'
            description = '{summary}'
            authors = [
                {{name = '{author}', email = '{author_email}'}}
            ]

            dependencies = [
                # Your package dependencies go here.
                # Don't include "ReviewBoard" in this list.
            ]

            classifiers = [
                # For a full list of package classifiers, see
                # https://pypi.python.org/pypi?%3Aaction=list_classifiers

                'Development Status :: 3 - Alpha',
                'Environment :: Web Framework',
                'Framework :: Review Board',
                'Operating System :: OS Independent',
                'Programming Language :: Python',
            ]


            [project.entry-points."reviewboard.extensions"]
            {package_name} = '{package_name}.extension:{class_name}'


            [tool.setuptools.packages.find]
            where = ['.']
            namespaces = false


            [tool.pytest.ini_options]
            DJANGO_SETTINGS_MODULE = "reviewboard.settings"
            django_debug_mode = false

            python_files = ["tests.py", "test_*.py"]
            python_classes = ["*Tests"]
            python_functions = ["test_*"]
            pythonpath = "."
            testpaths = ["{package_name}"]

            env = [
                "RB_RUNNING_TESTS=1",
                "RBSSH_STORAGE_BACKEND=reviewboard.ssh.storage.FileSSHStorage",
            ]

            addopts = ["--reuse-db"]

            required_plugins = [
                "pytest-django",
                "pytest-env",
            ]
        """

    def _create_conftest_py(self) -> str:
        """Return the content for a conftest.py file.

        Returns:
            str:
            The resulting content for the file.
        """
        return """
            pytest_plugins = ['reviewboard.testing.pytest_fixtures']
        """

    def _create_extension_py(
        self,
        name: str,
        package_name: str,
        class_name: str,
        summary: str,
        configurable: bool,
        has_static_media: bool,
    ) -> str:
        """Create the content for an extension.py file.

        Args:
            name (str):
                The name of the extension.

            package_name (str):
                The name of the package.

            class_name (str):
                The name of the extension class.

            summary (str):
                A summary of the extension.

            configurable (bool):
                Whether the package is set to be configurable.

            has_static_media (bool):
                Whether the package is set to have static media files.

        Returns:
            str:
            The resulting content for the file.
        """
        extension_docs_url = f'{EXTENSION_MANUAL_URL}extensions/'

        static_docs_url = f'{extension_docs_url}static-files/'
        static_media_content = f"""
                # You can create a list of CSS bundles to compile and ship
                # with your extension. These can include both *.css and
                # *.less (http://lesscss.org/) files. See
                # {static_docs_url}
                css_bundles = {{
                    # 'my-bundle-name': {{
                    #     'source_filenames': [
                    #         'css/style.less',
                    #     ],
                    #     'apply_to': ['my-view-url-name'],
                    # }},
                }}

                # JavaScript bundles are also supported. These support
                # standard .js (plain JavaScript), .es6.js (transpiled ES6+
                # JavaScript), and .ts (TypeScript) files.
                js_bundles = {{
                    # 'my-bundle-name': {{
                    #     'source_filenames': [
                    #         'js/script.es6.js',
                    #         'js/another-script.js',
                    #     ],
                    # }},
                }}
        """

        configure_docs_url = f'{extension_docs_url}configuration/'
        configuration_content = f"""
                # Default values for any configuration settings for your
                # extension.
                default_settings = {{
                    # 'my_field_1': 'my default value',
                    # 'my_field_2': False,
                }}

                # Set is_configurable and define an admin_urls.py to add
                # a standard configuration page for your extension.
                # See {configure_docs_url}
                is_configurable = True
        """

        metadata_name = repr(name)
        metadata_summary = repr(summary or 'REPLACE ME')
        hooks_docs_url = f'{extension_docs_url}#python-extension-hooks'
        extra_class_content = ''.join(
            extra_content
            for extra_content, should_add in (
                (configuration_content, configurable),
                (static_media_content, has_static_media))
            if should_add
        )

        return f'''
            """{name} for Review Board."""

            from django.utils.translation import gettext_lazy as _
            from reviewboard.extensions.base import Extension


            class {class_name}(Extension):
                """Internal description for your extension here."""

                metadata = {{
                    'Name': _({metadata_name}),
                    'Summary': _({metadata_summary}),
                }}
            {extra_class_content}
                def initialize(self) -> None:
                    """Initialize the extension."""
                    # Set up any hooks your extension needs here. See
                    # {hooks_docs_url}
        '''

    def _create_admin_urls_py(
        self,
        package_name: str,
        class_name: str,
        form_class_name: str,
    ) -> str:
        """Create the content for an admin_urls.py file.

        Args:
            package_name (str):
                The name of the package.

            class_name (str):
                The name of the extension class.

            form_class_name (str):
                The name of the extension settings form class.

        Returns:
            str:
            The resulting content for the file.
        """
        return f'''
            """Administration and configuration URLs for the extension."""

            from django.urls import path
            from reviewboard.extensions.views import configure_extension

            from {package_name}.extension import {class_name}
            from {package_name}.forms import {form_class_name}


            urlpatterns = [
                path(
                    '',
                    configure_extension,
                    {{
                        'ext_class': {class_name},
                        'form_class': {form_class_name},
                    }}),
            ]
        '''

    def _create_forms_py(
        self,
        form_class_name: str,
    ) -> str:
        """Create the content for a forms.py file.

        Args:
            form_class_name (str):
                The name of the extension settings form class.

        Returns:
            str:
            The resulting content for the file.
        """
        return f'''
            """Configuration forms for the extension."""

            from django import forms
            from djblets.extensions.forms import SettingsForm


            class {form_class_name}(SettingsForm):
                """Settings form for the extension."""

                # Add your fields here:
                # my_field_1 = forms.CharField()
                # my_field_2 = forms.BooleanField()
        '''


class RBExt:
    """Command line tool for helping develop Review Board extensions.

    This tool provides subcommands useful for extension developers. It
    currently provides:

    * ``test``: Runs an extension's test suite.
    """

    COMMANDS = [
        CreateCommand(),
        TestCommand(),
    ]

    def run(
        self,
        argv: Sequence[str],
    ) -> int:
        """Run an rbext command with the provided arguments.

        During the duration of the run, :py:data:`sys.argv` will be set to
        the provided arguments.

        Args:
            argv (list of str):
                The command line arguments passed to the command. This should
                not include the executable name as the first element.

        Returns:
            int:
            The command's exit code.
        """
        global console

        console = init_console()

        command, options = self.parse_options(argv)

        # We call out to things like the test runner, which expect to operate
        # off of sys.argv. We want to simulate that now that we've parsed
        # options. We'll restore sys.argv after the command finishes.
        old_argv = sys.argv
        sys.argv = argv

        try:
            return command.run(options)
        except Exception as e:
            logger.exception('Unexpected exception when running command '
                             '"%s": %s',
                             command.name, e)
            return 1
        finally:
            sys.argv = old_argv

    def parse_options(
        self,
        argv: Sequence[str],
    ) -> tuple[BaseCommand, argparse.Namespace]:
        """Parse arguments for the command.

        Args:
            argv (list of str):
                The arguments provided on the command line.

        Returns:
            tuple:
            A 2-tuple of:

            Tuple:
                0 (BaseCommand):
                    The command to run.

                1 (argparse.Namespace):
                    The parsed arguments.
        """
        parser = argparse.ArgumentParser(
            prog='rbext',
            usage='%(prog)s <command>',
            formatter_class=HelpFormatter,
            description=(
                'rbext helps create initial source code trees for extensions '
                'and helps run extension test suites within a '
                'pre-established Review Board development environment.\n'
                '\n'
                'To get help on an individual command, run:\n'
                '\n'
                '  rbext <command> --help'
            ))
        parser.add_argument(
            '--version',
            action=RBProgVersionAction)

        subparsers = parser.add_subparsers(
            title='Commands',
            dest='command',
            description=(
                'To get additional help for these commands, run: '
                'rb-site <command> --help'
            ))

        commands = sorted(self.COMMANDS, key=lambda cmd: cmd.name)
        command_map = {}

        for command in commands:
            command_map[command.name] = command

            subparser = subparsers.add_parser(
                command.name,
                formatter_class=HelpFormatter,
                prog=f'{parser.prog} {command.name}',
                description=command.description_text,
                help=command.help_summary)

            subparser.add_argument(
                '-d',
                '--debug',
                action='store_true',
                dest='debug',
                default=False,
                help='Display debug output.')
            subparser.add_argument(
                '-s',
                '--settings-file',
                dest='settings_file',
                default=None,
                help='test_settings.py file to use for any custom settings.')

            command.add_options(subparser)

        # Prevent the '--' and anything after it from being parsed, so the
        # command can work with it.
        try:
            i = argv.index('--')
            argv = argv[:i]
        except ValueError:
            # The "--" isn't in the list anywhere.
            pass

        options = parser.parse_args(argv)

        if not options.command:
            parser.print_help()
            sys.exit(1)

        return command_map[options.command], options


def main() -> None:
    """Run rbext.

    This is used by the Python EntryPoint to run rbext. It will pass in any
    arguments found on the command line and exit with the correct error code.
    """
    sys.exit(RBExt().run(sys.argv[1:]))



"""Unit tests for reviewboard.search.search_backends.elasticsearch.

Version Added:
    5.0.4
"""

from __future__ import annotations

import kgb
from django.core.exceptions import ValidationError

from reviewboard.search.search_backends.base import SearchBackend
from reviewboard.search.search_backends.elasticsearch import (
    ElasticsearchBackend)
from reviewboard.testing import TestCase


class ElasticsearchBackendTests(kgb.SpyAgency, TestCase):
    """Unit tests for ElasticsearchBackend.

    Version Added:
        5.0.4
    """

    def setUp(self):
        # Don't let the default validation logic happen, since it'll be more
        # dependent on package/version availability.
        self.spy_on(SearchBackend.validate,
                    owner=SearchBackend,
                    call_original=False)

    def test_validate(self):
        """Testing ElasticsearchBackend.validate"""
        backend = ElasticsearchBackend()
        backend._ES_VERSION_SUPPORTED = True
        backend._es_version = [7, 0, 0]

        # This should not raise an exception.
        backend.validate(configuration={})

    def test_validate_with_no_dep_installed(self):
        """Testing ElasticsearchBackend.validate with elasticsearch dependency
        not installed
        """
        backend = ElasticsearchBackend()
        backend._ES_VERSION_SUPPORTED = False
        backend._es_version = None

        message = (
            'You need to install a supported version of the elasticsearch '
            'module. Compatible packages are: ReviewBoard[elasticsearch7], '
            'ReviewBoard[elasticsearch5], ReviewBoard[elasticsearch2]'
            'ReviewBoard[elasticsearch1]'
        )

        with self.assertRaises(ValidationError) as e:
            backend.validate(configuration={})

            self.assertEqual(e.errors, [message])

    def test_validate_with_incompatible_version_installed(self):
        """Testing ElasticsearchBackend.validate with incompatible
        elasticsearch dependency version
        """
        backend = ElasticsearchBackend()
        backend._ES_VERSION_SUPPORTED = False
        backend._es_version = [4, 1, 2]

        message = (
            'You need to install a supported version of the elasticsearch '
            'module. version 4.1.2 is not supported. Compatible packages are: '
            'ReviewBoard[elasticsearch7], ReviewBoard[elasticsearch5], '
            'ReviewBoard[elasticsearch2], ReviewBoard[elasticsearch1]'
        )

        with self.assertRaises(ValidationError) as e:
            backend.validate(configuration={})

            self.assertEqual(e.errors, [message])
